"""
Point process module for expanding trees to a region of interest (ROI) and
generating random tree locations based on a specified point process.
"""

# Core imports
from __future__ import annotations

# External imports
import numpy as np
import pandas as pd
import geopandas as gpd
from numpy import ndarray
from pandas import DataFrame
from geopandas import GeoDataFrame
from scipy.interpolate import griddata


class InhomogeneousPoissonProcess:
    def simulate(
        self,
        roi: GeoDataFrame,
        trees: DataFrame,
        plots: GeoDataFrame,
        intensity_resolution: int = 15,
        intensity_interpolation_method: str = "linear",
        seed: int = None,
    ) -> GeoDataFrame:
        """
        Generates random tree locations based on an inhomogeneous Poisson
        point process.
        """
        self._set_seed(seed)
        grid_x, grid_y = self._create_structured_coords_grid(roi, intensity_resolution)
        tree_density_grid = self._interpolate_tree_density_to_grid(
            trees,
            plots,
            grid_x,
            grid_y,
            intensity_resolution,
            intensity_interpolation_method,
        )
        plot_id_grid = self._interpolate_plot_id_to_grid(trees, plots, grid_x, grid_y)
        tree_count_grid = self._generate_tree_counts(tree_density_grid)
        tree_locations = self._generate_tree_locations_in_grid(
            tree_count_grid,
            plot_id_grid,
            intensity_resolution,
            grid_x,
            grid_y,
        )
        sampled_trees = self._sample_trees_by_location(trees, tree_locations)
        trees_in_bounds = self._drop_trees_outside_roi_bounds(sampled_trees, roi)
        return self._convert_to_geodataframe(trees_in_bounds, roi.crs)

    @staticmethod
    def _set_seed(seed):
        """Sets the seed for the random number generator."""
        seed = np.random.randint(0, 1000000) if seed is None else seed
        np.random.seed(seed)

    @staticmethod
    def _create_structured_coords_grid(roi, resolution) -> tuple[ndarray, ndarray]:
        """
        Creates a structured grid of cell-centered coordinates for the ROI.
        The resolution parameter specifies the size of each cell in the grid.
        """
        west, south, east, north = roi.total_bounds
        x = np.arange(west + resolution / 2, east, resolution)
        y = np.arange(north - resolution / 2, south, -resolution)
        xx, yy = np.meshgrid(x, y)
        return xx, yy

    def _interpolate_tree_density_to_grid(
        self,
        trees,
        plots,
        grid_x,
        grid_y,
        cell_resolution,
        interpolation_method="linear",
    ) -> ndarray:
        """
        Interpolates tree density to a structured grid of cells. Tree density
        is calculated for each plot and then interpolated to the grid using
        the specified interpolation method, typically "linear" or "cubic".
        """
        plots_with_data_col = self._calculate_all_plots_tree_density(plots, trees)
        data_to_interpolate = plots_with_data_col["TPA"] * (cell_resolution**2)
        interpolated_plot_data = self._interpolate_data_to_grid(
            plots_with_data_col,
            data_to_interpolate,
            grid_x,
            grid_y,
            interpolation_method,
        )
        return interpolated_plot_data

    def _interpolate_plot_id_to_grid(self, trees, plots, grid_x, grid_y) -> ndarray:
        """
        Interpolates the plot IDs of plots containing trees to a structured
        grid of cells. The plot ID is a categorical value that is used to
        assign trees to plots. This function effectively creates a
        discretized Veroni diagram of the plots containing trees.
        """
        plots_with_data_col = self._calculate_occupied_plots_tree_density(plots, trees)
        data_to_interpolate = plots_with_data_col["PLOT_ID"]
        interpolated_plot_data = self._interpolate_data_to_grid(
            plots_with_data_col, data_to_interpolate, grid_x, grid_y, "nearest"
        )
        return interpolated_plot_data

    def _calculate_all_plots_tree_density(self, plots, trees) -> GeoDataFrame:
        """
        Calculates the tree density, stored in the TPA column, for all
        plots in the ROI. Returns a DataFrame of all plots with a TPA
        column containing the tree density for each plot.
        """
        return self._calculate_per_plot_tree_density(plots, trees, "left")

    def _calculate_occupied_plots_tree_density(self, plots, trees) -> GeoDataFrame:
        """
        Calculates the tree density, stored in the TPA column, for plots
        containing trees in the ROI. Returns a DataFrame of plots containing
        trees with a TPA column containing the tree density for each
        plot.
        """
        return self._calculate_per_plot_tree_density(plots, trees, "right")

    @staticmethod
    def _calculate_per_plot_tree_density(plots, trees, merge_type) -> GeoDataFrame:
        """
        Calculate the tree density for a specific grouping of plots based on
        the merge type. if the merge type is "left", the tree density is
        calculated for all plots in the ROI. If the merge type is "right", the
        tree density is calculated only for plots containing trees in the ROI.
        Returns a DataFrame of plots with a TPA column containing the
        tree density for each plot.
        """
        sum_by_plot = trees.groupby("PLOT_ID")["TPA"].sum().reset_index()
        merged_plots = plots.merge(sum_by_plot, on="PLOT_ID", how=merge_type)
        return merged_plots.fillna(0)

    @staticmethod
    def _interpolate_data_to_grid(plots, data, grid_x, grid_y, method) -> ndarray:
        """
        Interpolate unstructured plot data to a structured grid.
        """
        # If there is no data to interpolate, return a grid of zeros.
        # This can happen if there are no trees in the ROI.
        if len(data) == 0:
            return np.zeros(grid_x.shape)

        interpolated_grid = griddata(
            (plots.geometry.x, plots.geometry.y),
            data,
            (grid_x, grid_y),
            method=method,
        )
        interpolated_grid = np.nan_to_num(interpolated_grid, nan=0)
        interpolated_grid[interpolated_grid < 0] = 0

        return interpolated_grid

    @staticmethod
    def _generate_tree_counts(intensity_grid) -> ndarray:
        """
        Draws a random number of trees from a Poisson distribution for each
        cell in the intensity grid. The intensity grid is a 2D array where
        each cell represents a plot and the value in the cell represents the
        tree density of the plot.
        """
        return np.random.poisson(intensity_grid)

    def _generate_tree_locations_in_grid(
        self, count_grid, plot_grid, grid_resolution, grid_x, grid_y
    ):
        """
        Generates tree locations in a grid of plots.
        """
        flat_tree_indices = self._get_flattened_tree_indices(count_grid)
        cell_i, cell_j = self._get_grid_cell_indices(flat_tree_indices, count_grid)
        x_coords, y_coords = self._calculate_tree_coordinates(
            cell_i, cell_j, grid_resolution, grid_x, grid_y
        )
        return pd.DataFrame(
            {
                "X": x_coords,
                "Y": y_coords,
                "PLOT_ID": plot_grid.ravel()[flat_tree_indices],
            }
        )

    @staticmethod
    def _get_flattened_tree_indices(count_grid: ndarray) -> ndarray:
        """
        Generate a list of indices representing the locations of trees in a
        grid.

        This function takes a 2D grid (a numpy array) as input, where each
        cell in the grid represents a plot of land and the value in the cell
        represents the number of trees in that plot. The function uses the
        `numpy.repeat` function to repeat each index in the flattened grid
        according to the number of trees in the corresponding plot.

        The output of this function is a 1D numpy array where each element is
        an index of a cell in the flattened grid, and each index is repeated
        a number of times equal to the number of trees in the corresponding
        plot. This array of indices can then be used to generate the
        locations of individual trees within each plot.
        """
        return np.repeat(np.arange(count_grid.size), count_grid.ravel())

    @staticmethod
    def _get_grid_cell_indices(tree_indices, count_grid) -> tuple[ndarray, ndarray]:
        """
        Converts the flattened tree indices into 2D cell indices. Returns two
        1D arrays containing the row and column indices of each tree.

        This method takes a 1D array of tree indices and the 2D count grid as
        input. It uses the `numpy.unravel_index` function to convert the
        flattened tree indices into 2D cell indices that correspond to their
        original locations in the count grid.
        """
        return np.unravel_index(tree_indices, count_grid.shape)

    def _calculate_tree_coordinates(
        self, cell_i, cell_j, grid_resolution, grid_x, grid_y
    ) -> tuple[ndarray, ndarray]:
        """
        Calculates the x and y coordinates of trees within each plot in the
        grid. Returns two 1D arrays containing the x and y coordinates of each
        tree.

        This function takes the row and column indices of each tree's
        location in the grid, the resolution of the grid, and the x and y
        coordinates of the grid as inputs. It generates random offsets for
        each tree within its respective plot, ensuring that the tree's
        location is within the plot's boundaries. The offsets are generated
        uniformly at random within the range of [-grid_resolution / 2,
        grid_resolution / 2]. These offsets are then added to the x and y
        coordinates of the plot's center to get the final coordinates of each
        tree.
        """
        random_offsets_x, random_offsets_y = self._generate_random_offsets(
            len(cell_i), grid_resolution
        )
        x_coords = grid_x[cell_i, cell_j] + random_offsets_x
        y_coords = grid_y[cell_i, cell_j] + random_offsets_y
        return x_coords, y_coords

    @staticmethod
    def _generate_random_offsets(num_indices, grid_resolution):
        """
        Generates a random offset for each tree centered at 0. The offsets are
        generated uniformly at random within the range of
        [-grid_resolution / 2, grid_resolution / 2].
        """
        return (
            np.random.uniform(-grid_resolution / 2, grid_resolution / 2, num_indices),
            np.random.uniform(-grid_resolution / 2, grid_resolution / 2, num_indices),
        )

    def _sample_trees_by_location(self, trees, tree_locations):
        """
        This function first checks if the trees belong to plots. If they do,
        it counts the number of trees per plot, samples trees based on plot
        intensity, and sorts the tree locations by PlotID.

        If the trees do not belong to plots, it samples the trees without
        considering plots.

        The function then assigns the generated tree locations to the sampled
        trees. Finally, it converts the sampled trees to a GeoDataFrame with
        spatial coordinates and returns it.
        """
        num_trees_per_plot = self._count_trees_per_plot(trees, tree_locations)
        sampled_trees = self._sample_trees_within_plots(trees, num_trees_per_plot)
        sampled_trees = self._assign_tree_locations(sampled_trees, tree_locations)
        return sampled_trees

    @staticmethod
    def _count_trees_per_plot(trees, tree_locations):
        """
        Count the number of trees per plot.
        """
        tree_counts = tree_locations.value_counts("PLOT_ID", dropna=False)
        all_plot_ids = trees["PLOT_ID"].unique()
        return tree_counts.reindex(all_plot_ids, fill_value=0)

    @staticmethod
    def _sample_trees_within_plots(trees, num_trees_per_plot):
        """
        Sample trees based on plot intensity, with handling for PLOT_IDs not in num_trees_per_plot.
        """

        # Define the sampling function
        def sample_trees_for_plot(plot):
            return plot.sample(
                n=num_trees_per_plot.loc[plot.name], weights=plot["TPA"], replace=True
            )

        # Apply the sampling function to each plot
        sampled_trees = trees.groupby("PLOT_ID").apply(sample_trees_for_plot)

        return sampled_trees.reset_index(drop=True)

    @staticmethod
    def _assign_tree_locations(sampled_trees, tree_locations):
        """
        Assign tree locations to sampled trees.
        """
        # Align sampled_trees and tree_locations by sorting by PLOT_ID
        tree_locations = tree_locations.sort_values("PLOT_ID")
        sampled_trees = sampled_trees.sort_values("PLOT_ID")

        # Assign tree locations to sampled trees
        sampled_trees["X"] = tree_locations["X"].values
        sampled_trees["Y"] = tree_locations["Y"].values

        return sampled_trees

    @staticmethod
    def _drop_trees_outside_roi_bounds(df, roi):
        """
        Drop trees that are outside the total bounds of the ROI.
        """
        minx, miny, maxx, maxy = roi.total_bounds
        df.drop(df[(df["Y"] < miny) | (df["Y"] > maxy)].index, inplace=True)
        df.drop(df[(df["X"] < minx) | (df["X"] > maxx)].index, inplace=True)
        return df

    @staticmethod
    def _convert_to_geodataframe(sampled_trees, crs):
        """
        Convert sampled trees to a GeoDataFrame with spatial coordinates.
        """
        return gpd.GeoDataFrame(
            sampled_trees,
            geometry=gpd.points_from_xy(sampled_trees["X"], sampled_trees["Y"]),
            crs=crs,
        )


def run_point_process(process_type: str, roi, trees, **kwargs):
    if process_type == "inhomogeneous_poisson":
        point_process = InhomogeneousPoissonProcess()
    else:
        raise ValueError(f"Invalid point process type: {process_type}")
    return point_process.simulate(roi, trees, **kwargs)



# Core imports
import math
import random
from pathlib import Path

# Internal imports
from tests.utils import make_random_tree
from fastfuels_core.trees import Tree
from fastfuels_core.voxelization import (
    _get_horizontal_tree_coords,
    _get_vertical_tree_coords,
    _resample_coords_grid_to_subgrid,
    _compute_circle_segment_area,
    _calculate_case_1_area,
    _calculate_case_3_area,
    _calculate_case_9_area,
    _calculate_case_11_area,
    _encode_corners,
    _compute_intersection_area_by_case,
    _sum_area_along_axis,
    discretize_crown_profile,
    _compute_horizontal_probability,
    _compute_vertical_probability,
    _compute_joint_probability,
    _sample_voxels_from_probability_grid,
    voxelize_tree,
)

# External imports
import pytest
import numpy as np
import matplotlib.pyplot as plt
from shapely.geometry import Polygon, Point

TEST_PATH = Path(__file__).parent
FIGURES_PATH = TEST_PATH / "figures"


class TestGetHorizontalTreeCoords:
    def test_positive_radius(self):
        step = 1
        radius = 5
        pos = 0
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 13
        assert result[0] == -radius - step
        assert result[1] == -radius
        assert result[6] == 0
        assert result[-2] == radius
        assert result[-1] == radius + step

    def test_zero_radius(self):
        step = 1
        radius = 0
        pos = 0
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 3
        assert result[0] == -1
        assert result[0] + step == -radius
        assert result[1] == 0
        assert result[-1] == 1
        assert result[-1] - step == radius

    def test_subgrid_radius(self):
        step = 1
        radius = 0.1
        pos = 0
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 3
        assert result[0] == -1
        assert result[1] == 0
        assert result[-1] == 1

    def test_negative_radius(self):
        step = 1
        radius = -5
        pos = 0
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 13
        assert result[0] == -6
        assert result[0] + step == radius
        assert result[6] == 0
        assert result[-1] == 6
        assert result[-1] - step == -radius

    def test_fractional_step(self):
        step = 0.5
        radius = 5
        pos = 0
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 23
        assert result[0] == -5.5
        assert result[0] + step == -radius
        assert result[11] == 0
        assert result[-1] == 5.5
        assert result[-1] - step == radius

    def test_fine_step(self):
        step = 0.1
        radius = 5
        pos = 0
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 103
        assert np.allclose(result[0], -5 - step)
        assert np.allclose(result[1], -radius)
        assert result[51] == 0
        assert np.allclose(result[-2], radius)
        assert np.allclose(result[-1], radius + step)

    def test_zero_step(self):
        step = 0
        radius = 5
        pos = 0
        with pytest.raises(ZeroDivisionError):
            _get_horizontal_tree_coords(step, radius, pos)

    def test_non_zero_pos(self):
        step = 1
        radius = 5
        pos = 2
        result = _get_horizontal_tree_coords(step, radius, pos)
        assert isinstance(result, np.ndarray)
        assert len(result) == 13
        assert result[0] == -4
        assert result[0] + step == -radius + pos
        assert result[6] == 2
        assert result[-1] == 8
        assert result[-1] - step == radius + pos


class TestGetVerticalTreeCoords:
    def test_positive_height(self):
        step = 1
        tree_height = 5
        crown_base_height = 0
        result = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(result, np.ndarray)
        assert len(result) == 6
        assert result[0] == crown_base_height
        assert result[-1] == tree_height

    def test_fractional_height(self):
        step = 1
        tree_height = 10.1
        crown_base_height = 5
        result = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(result, np.ndarray)
        assert len(result) == 7
        assert result[0] == crown_base_height
        assert result[-1] == math.ceil(tree_height)

    def test_zero_height(self):
        step = 1
        tree_height = 0
        crown_base_height = 0
        result = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(result, np.ndarray)
        assert len(result) == 1
        assert result[0] == tree_height

    def test_negative_height(self):
        step = 1
        tree_height = -5
        crown_base_height = 0
        res = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(res, np.ndarray)
        assert len(res) == 0

    def test_fractional_step(self):
        step = 0.5
        tree_height = 5
        crown_base_height = 0
        result = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(result, np.ndarray)
        assert len(result) == 11
        assert result[0] == crown_base_height
        assert result[-1] == tree_height

    def test_fine_step(self):
        step = 0.1
        tree_height = 10.1
        crown_base_height = 5
        result = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(result, np.ndarray)
        assert len(result) == 52
        assert result[0] == crown_base_height
        assert np.allclose(result[-1], tree_height)

    def test_zero_step(self):
        step = 0
        tree_height = 5
        crown_base_height = 0
        with pytest.raises(ZeroDivisionError):
            _get_vertical_tree_coords(step, tree_height, crown_base_height)

    def test_fractional_cbh(self):
        step = 1
        tree_height = 20.3643
        crown_base_height = 5.891270987
        result = _get_vertical_tree_coords(step, tree_height, crown_base_height)
        assert isinstance(result, np.ndarray)
        assert len(result) == 16
        assert result[0] == crown_base_height
        assert result[-1] == 20.891270987


class TestResampleCoordsGridToSubgrid:
    def test_resample(self):
        original_pts = np.array([0.5])
        original_spacing = 1
        new_spacing = 0.5
        expected_pts_1 = np.array([0.25, 0.75])
        resampled_pts_1 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_1, expected_pts_1)

        original_pts = np.array([0.5])
        original_spacing = 1
        new_spacing = 0.25
        expected_pts_2 = np.array([0.125, 0.375, 0.625, 0.875])
        resampled_pts_2 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_2, expected_pts_2)

        original_pts = np.array([0.5])
        original_spacing = 1
        new_spacing = 0.1
        expected_pts_3 = np.array(
            [0.05, 0.15, 0.25, 0.35, 0.45, 0.55, 0.65, 0.75, 0.85, 0.95]
        )
        resampled_pts_3 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_3, expected_pts_3)

        original_pts = np.array([0.5, 1.5])
        original_spacing = 1
        new_spacing = 0.5
        expected_pts_4 = np.array([0.25, 0.75, 1.25, 1.75])
        resampled_pts_4 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_4, expected_pts_4)

        original_pts = np.array([0.5, 1.5, 2.5])
        original_spacing = 1
        new_spacing = 0.5
        expected_pts_5 = np.array([0.25, 0.75, 1.25, 1.75, 2.25, 2.75])
        resampled_pts_5 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_5, expected_pts_5)

        original_pts = np.array([0.5, 1.5, 2.5])
        original_spacing = 1
        new_spacing = 0.25
        expected_pts_6 = np.array(
            [
                0.125,
                0.375,
                0.625,
                0.875,
                1.125,
                1.375,
                1.625,
                1.875,
                2.125,
                2.375,
                2.625,
                2.875,
            ]
        )
        resampled_pts_6 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_6, expected_pts_6)

        original_pts = np.array([0.5, 1.5, 2.5])
        original_spacing = 1
        new_spacing = 0.1
        expected_pts_7 = np.array(
            [
                0.05,
                0.15,
                0.25,
                0.35,
                0.45,
                0.55,
                0.65,
                0.75,
                0.85,
                0.95,
                1.05,
                1.15,
                1.25,
                1.35,
                1.45,
                1.55,
                1.65,
                1.75,
                1.85,
                1.95,
                2.05,
                2.15,
                2.25,
                2.35,
                2.45,
                2.55,
                2.65,
                2.75,
                2.85,
                2.95,
            ]
        )
        resampled_pts_7 = _resample_coords_grid_to_subgrid(
            original_pts, original_spacing, new_spacing
        )
        assert np.allclose(resampled_pts_7, expected_pts_7)

    def test_same_resolution_shape(self):
        # Test 1m resolution
        grid = np.arange(0.5, 100.5, 1)
        grid_spacing = 1
        new_spacing = 1
        resampled_grid = _resample_coords_grid_to_subgrid(
            grid, grid_spacing, new_spacing
        )
        assert resampled_grid.shape == grid.shape

        # Test 0.5m resolution
        grid = np.arange(0.25, 100.25, 0.5)
        grid_spacing = 0.5
        new_spacing = 0.5
        resampled_grid = _resample_coords_grid_to_subgrid(
            grid, grid_spacing, new_spacing
        )
        assert resampled_grid.shape == grid.shape

        # Test 0.25m resolution
        grid = np.arange(0.125, 100.125, 0.25)
        grid_spacing = 0.25
        new_spacing = 0.25
        resampled_grid = _resample_coords_grid_to_subgrid(
            grid, grid_spacing, new_spacing
        )
        assert resampled_grid.shape == grid.shape

        # Test 0.1m resolution
        grid = np.arange(0.05, 100.05, 0.1)
        grid_spacing = 0.1
        new_spacing = 0.1
        resampled_grid = _resample_coords_grid_to_subgrid(
            grid, grid_spacing, new_spacing
        )
        assert resampled_grid.shape == grid.shape


class TestComputeCircleSegmentArea:
    def test_zero_radius(self):
        # Test case where radius is zero
        csa = _compute_circle_segment_area(0, 0.5, 1, 0.5, 0)
        assert csa == 0

    def test_zero_chord_length(self):
        # Test case where chord length is zero (points are the same)
        csa = _compute_circle_segment_area(1, 1, 1, 1, 0.5)
        assert csa == 0


class TestCalculateCase1Area:
    save_fig = False
    show_fig = False

    def test_with_floats(self):
        bottom = 0
        left = 0
        radius = 0.75
        intersection_area = _calculate_case_1_area(left, bottom, radius)
        assert isinstance(intersection_area, float)
        assert 0 < intersection_area < 1

    def test_with_numpy(self):
        bottom = np.array([0])
        left = np.array([0])
        radius = np.array([0.75])
        intersection_area = _calculate_case_1_area(left, bottom, radius)
        assert isinstance(intersection_area, np.ndarray)
        assert intersection_area.size == 1
        assert 0 < intersection_area[0] < 1

    def test_compare_with_shapely_exact(self):
        length = 1
        radius = np.array([1e-4, 0.15, 0.25, 0.5, 0.75, 1.0])
        x_left = np.zeros_like(radius)
        y_bottom = np.zeros_like(radius)
        intersection = _calculate_case_1_area(x_left, y_bottom, radius, exact=True)

        bottom_left = Point(0, 0)
        bottom_right = Point(length, 0)
        top_right = Point(length, length)
        top_left = Point(0, length)
        cell = Polygon([bottom_left, bottom_right, top_right, top_left])
        circle = Point(0, 0).buffer(radius, quad_segs=2**16)
        intersection_shapely = cell.intersection(circle)

        assert len(intersection) == len(intersection_shapely)
        assert np.allclose(intersection[0], intersection_shapely[0].area)
        assert np.allclose(intersection[1], intersection_shapely[1].area)
        assert np.allclose(intersection[2], intersection_shapely[2].area)
        assert np.allclose(intersection[3], intersection_shapely[3].area)
        assert np.allclose(intersection[4], intersection_shapely[4].area)
        assert np.allclose(intersection[5], intersection_shapely[5].area)

        # plot the cell and the circle
        fig, ax = plt.subplots(3, 2, figsize=(6, 8))
        for row in range(3):
            for col in range(2):
                i = row * 2 + col
                ax[row, col].plot(*cell.exterior.xy, lw=5, label="cell")
                ax[row, col].plot(*circle[i].exterior.xy, lw=3, c="g", label="circle")
                ax[row, col].plot(
                    *intersection_shapely[i].exterior.xy,
                    "--",
                    lw=2,
                    c="r",
                    label="intersection"
                )
                ax[row, col].set_xlim(-radius[i] / 8, radius[i] + radius[i] / 8)
                ax[row, col].set_ylim(-radius[i] / 8, radius[i] + radius[i] / 8)
                ax[row, col].set_ylabel("y (m)")
                ax[row, col].set_xlabel("x (m)")
                ax[row, col].set_aspect("equal")
                # ax[row, col].legend(loc="upper right")
        # plt.legend()
        plt.tight_layout()
        if self.save_fig:
            plt.savefig(FIGURES_PATH / "test_calculate_case_1_area.png")
        if self.show_fig:
            plt.show()


class TestCalculateCase3Area:
    save_fig = False
    show_fig = False

    def test_with_floats(self):
        left = -0.5
        right = 0.5
        bottom = 0.5
        length = 1
        radius = 1.25
        intersection_area = _calculate_case_3_area(left, right, bottom, radius, length)
        assert isinstance(intersection_area, float)
        assert 0 < intersection_area < 1

    def test_with_numpy(self):
        left = np.array([-0.5])
        right = np.array([0.5])
        bottom = np.array([0.5])
        length = np.array([1])
        radius = np.array([1.25])
        intersection_area = _calculate_case_3_area(left, right, bottom, radius, length)
        assert isinstance(intersection_area, np.ndarray)
        assert intersection_area.size == 1
        assert 0 < intersection_area[0] < 1

    def test_compare_with_shapely_exact(self):
        length = 1
        radius = np.array([np.sqrt(0.5), 0.75, 1.0, 1.15, 1.35, 1.5])
        left = np.zeros_like(radius) - length / 2
        right = np.zeros_like(radius) + length / 2
        bottom = np.zeros_like(radius) + length / 2
        intersection = _calculate_case_3_area(
            left, right, bottom, radius, length, exact=True
        )

        bottom_left = Point(-length / 2, length / 2)
        bottom_right = Point(length / 2, length / 2)
        top_left = Point(-length / 2, 3 * length / 2)
        top_right = Point(length / 2, 3 * length / 2)
        cell = Polygon([bottom_left, bottom_right, top_right, top_left])
        circle = Point(0, 0).buffer(radius, quad_segs=2**16)
        intersection_shapely = cell.intersection(circle)

        assert len(intersection) == len(intersection_shapely)
        assert np.allclose(intersection[0], intersection_shapely[0].area)
        assert np.allclose(intersection[1], intersection_shapely[1].area)
        assert np.allclose(intersection[2], intersection_shapely[2].area)
        assert np.allclose(intersection[3], intersection_shapely[3].area)
        assert np.allclose(intersection[4], intersection_shapely[4].area)
        assert np.allclose(intersection[5], intersection_shapely[5].area)

        # plot the cell and the circle
        nrows = radius.size // 2
        ncols = 2
        fig, ax = plt.subplots(nrows, ncols)
        for row in range(nrows):
            for col in range(ncols):
                i = row * 2 + col
                ax[row, col].plot(*cell.exterior.xy, lw=5, label="cell")
                ax[row, col].plot(*circle[i].exterior.xy, lw=3, c="g", label="circle")
                ax[row, col].plot(
                    *intersection_shapely[i].exterior.xy,
                    "--",
                    lw=2,
                    c="r",
                    label="intersection"
                )
                ax[row, col].set_xlim(-length / 2 - 0.01, length / 2 + 0.01)
                ax[row, col].set_ylim(length / 2, 3 * length / 2)
                ax[row, col].set_ylabel("y (m)")
                ax[row, col].set_xlabel("x (m)")
                ax[row, col].set_aspect("equal")

        plt.tight_layout()
        if self.save_fig:
            plt.savefig(FIGURES_PATH / "test_calculate_case_3_area.png")
        if self.show_fig:
            plt.show()

    def test_compare_with_shapely_offset_cell_exact(self):
        length = 0.5
        radius = np.array([1.46, 1.5, 1.6, 1.75])
        left = np.zeros_like(radius) + length / 2
        right = np.zeros_like(radius) + 3 * length / 2
        bottom = np.zeros_like(radius) + 5 * length / 2
        top = np.zeros_like(radius) + 7 * length / 2
        intersection = _calculate_case_3_area(
            left, right, bottom, radius, length, exact=True
        )

        bottom_left = Point(left[0], bottom[0])
        bottom_right = Point(right[0], bottom[0])
        top_left = Point(left[0], top[0])
        top_right = Point(right[0], top[0])
        cell = Polygon([bottom_left, bottom_right, top_right, top_left])
        circle = Point(0.0, 0.0).buffer(radius, quad_segs=2**16)
        intersection_shapely = cell.intersection(circle)

        assert len(intersection) == len(intersection_shapely)
        assert np.allclose(intersection[0], intersection_shapely[0].area)
        assert np.allclose(intersection[1], intersection_shapely[1].area)
        assert np.allclose(intersection[2], intersection_shapely[2].area)
        assert np.allclose(intersection[3], intersection_shapely[3].area)

        # plot the cell and the circle
        nrows = radius.size // 2
        ncols = 2
        fig, ax = plt.subplots(nrows, ncols)
        for row in range(nrows):
            for col in range(ncols):
                i = row * 2 + col
                ax[row, col].plot(*cell.exterior.xy, lw=5, label="cell")
                ax[row, col].plot(*circle[i].exterior.xy, lw=3, c="g", label="circle")
                ax[row, col].plot(
                    *intersection_shapely[i].exterior.xy,
                    "--",
                    lw=2,
                    c="r",
                    label="intersection"
                )
                ax[row, col].set_xlim(left[0], right[0])
                ax[row, col].set_ylim(bottom[0], top[0])
                ax[row, col].set_ylabel("y (m)")
                ax[row, col].set_xlabel("x (m)")
                ax[row, col].set_aspect("equal")

        plt.tight_layout()
        if self.save_fig:
            plt.savefig(FIGURES_PATH / "test_calculate_case_3_area_offset.png")
        if self.show_fig:
            plt.show()


class TestCalculateCase9Area:
    save_fig = False
    show_fig = False

    def test_with_floats(self):
        left = 0.5
        bottom = -0.5
        top = 0.5
        length = 1
        radius = 1.25
        intersection_area = _calculate_case_9_area(top, bottom, left, radius, length)
        assert isinstance(intersection_area, float)
        assert 0 < intersection_area < 1

    def test_with_numpy(self):
        left = np.array([0.5])
        bottom = np.array([-0.5])
        top = np.array([0.5])
        length = np.array([1])
        radius = np.array([1.25])
        intersection_area = _calculate_case_9_area(top, bottom, left, radius, length)
        assert isinstance(intersection_area, np.ndarray)
        assert intersection_area.size == 1
        assert 0 < intersection_area[0] < 1

    def test_compare_with_shapely_exact(self):
        length = 1
        radius = np.array([np.sqrt(0.5), 0.75, 1.0, 1.15, 1.35, 1.5])
        top = np.zeros_like(radius) + length / 2
        bottom = np.zeros_like(radius) - length / 2
        left = np.zeros_like(radius) + length / 2
        right = np.zeros_like(radius) + 3 * length / 2
        intersection = _calculate_case_9_area(
            top, bottom, left, radius, length, exact=True
        )

        bottom_left = Point(left[0], bottom[0])
        bottom_right = Point(right[0], bottom[0])
        top_left = Point(left[0], top[0])
        top_right = Point(right[0], top[0])
        cell = Polygon([bottom_left, bottom_right, top_right, top_left])
        circle = Point(0, 0).buffer(radius, quad_segs=2**16)
        intersection_shapely = cell.intersection(circle)

        # assert len(intersection) == len(intersection_shapely)
        assert np.allclose(intersection[0], intersection_shapely[0].area)
        assert np.allclose(intersection[1], intersection_shapely[1].area)
        assert np.allclose(intersection[2], intersection_shapely[2].area)
        assert np.allclose(intersection[3], intersection_shapely[3].area)
        assert np.allclose(intersection[4], intersection_shapely[4].area)
        assert np.allclose(intersection[5], intersection_shapely[5].area)

        # plot the cell and the circle
        nrows = radius.size // 2
        ncols = 2
        fig, ax = plt.subplots(nrows, ncols)
        for row in range(nrows):
            for col in range(ncols):
                i = row * 2 + col
                ax[row, col].plot(*cell.exterior.xy, lw=5, label="cell")
                ax[row, col].plot(*circle[i].exterior.xy, lw=3, c="g", label="circle")
                ax[row, col].plot(
                    *intersection_shapely[i].exterior.xy,
                    "--",
                    lw=2,
                    c="r",
                    label="intersection"
                )
                ax[row, col].set_xlim(length / 2 - 0.01, 3 * length / 2 + 0.01)
                ax[row, col].set_ylim(-length / 2 - 0.01, length / 2 + 0.01)
                ax[row, col].set_ylabel("y (m)")
                ax[row, col].set_xlabel("x (m)")
                ax[row, col].set_aspect("equal")

        plt.tight_layout()
        if self.save_fig:
            plt.savefig(FIGURES_PATH / "test_calculate_case_9_area.png")
        if self.show_fig:
            plt.show()

    def test_compare_with_shapely_offset_cell_exact(self):
        length = 0.5
        radius = np.array([1.46, 1.5, 1.6, 1.75])
        bottom = np.zeros_like(radius) + length / 2
        top = np.zeros_like(radius) + 3 * length / 2
        left = np.zeros_like(radius) + 5 * length / 2
        right = np.zeros_like(radius) + 7 * length / 2
        intersection = _calculate_case_9_area(
            top, bottom, left, radius, length, exact=True
        )

        bottom_left = Point(left[0], bottom[0])
        bottom_right = Point(right[0], bottom[0])
        top_left = Point(left[0], top[0])
        top_right = Point(right[0], top[0])
        cell = Polygon([bottom_left, bottom_right, top_right, top_left])
        circle = Point(0.0, 0.0).buffer(radius, quad_segs=2**16)
        intersection_shapely = cell.intersection(circle)

        assert len(intersection) == len(intersection_shapely)
        assert np.allclose(intersection[0], intersection_shapely[0].area)
        assert np.allclose(intersection[1], intersection_shapely[1].area)
        assert np.allclose(intersection[2], intersection_shapely[2].area)
        assert np.allclose(intersection[3], intersection_shapely[3].area)

        # plot the cell and the circle
        nrows = radius.size // 2
        ncols = 2
        fig, ax = plt.subplots(nrows, ncols)
        for row in range(nrows):
            for col in range(ncols):
                i = row * 2 + col
                ax[row, col].plot(*cell.exterior.xy, lw=5, label="cell")
                ax[row, col].plot(*circle[i].exterior.xy, lw=3, c="g", label="circle")
                ax[row, col].plot(
                    *intersection_shapely[i].exterior.xy,
                    "--",
                    lw=2,
                    c="r",
                    label="intersection"
                )
                ax[row, col].set_xlim(left[0], right[0])
                ax[row, col].set_ylim(bottom[0], top[0])
                ax[row, col].set_ylabel("y (m)")
                ax[row, col].set_xlabel("x (m)")
                ax[row, col].set_aspect("equal")

        plt.tight_layout()
        if self.save_fig:
            plt.savefig(FIGURES_PATH / "test_calculate_case_9_area_offset.png")
        if self.show_fig:
            plt.show()


class TestCalculateCase11Area:
    save_fig = False
    show_fig = False

    def test_with_floats(self):
        right = 1.5
        top = 1.5
        radius = 1.75
        length = 1
        intersection_area = _calculate_case_11_area(top, right, radius, length)
        assert isinstance(intersection_area, float)
        assert 0 < intersection_area < 1

    def test_with_numpy(self):
        right = np.array([1.5])
        top = np.array([1.5])
        radius = np.array([1.75])
        length = np.array([1])
        intersection_area = _calculate_case_11_area(top, right, radius, length)
        assert isinstance(intersection_area, np.ndarray)
        assert intersection_area.size == 1
        assert 0 < intersection_area[0] < 1

    def test_compare_with_shapely_exact(self):
        length = 1
        radius = np.array([1.59, 1.6, 1.75, 1.9, 2.0, 2.1])
        left = np.zeros_like(radius) + length / 2
        right = np.zeros_like(radius) + 3 * length / 2
        bottom = np.zeros_like(radius) + length / 2
        top = np.zeros_like(radius) + 3 * length / 2
        intersection = _calculate_case_11_area(top, right, radius, length, exact=True)

        bottom_left = Point(left[0], bottom[0])
        bottom_right = Point(right[0], bottom[0])
        top_left = Point(left[0], top[0])
        top_right = Point(right[0], top[0])
        cell = Polygon([bottom_left, bottom_right, top_right, top_left])
        circle = Point(0, 0).buffer(radius, quad_segs=2**16)
        intersection_shapely = cell.intersection(circle)

        assert len(intersection) == len(intersection_shapely)
        assert np.allclose(intersection[0], intersection_shapely[0].area)
        assert np.allclose(intersection[1], intersection_shapely[1].area)
        assert np.allclose(intersection[2], intersection_shapely[2].area)
        assert np.allclose(intersection[3], intersection_shapely[3].area)
        assert np.allclose(intersection[4], intersection_shapely[4].area)
        assert np.allclose(intersection[5], intersection_shapely[5].area)

        # plot the cell and the circle
        nrows = radius.size // 2
        ncols = 2
        fig, ax = plt.subplots(nrows, ncols)
        for row in range(nrows):
            for col in range(ncols):
                i = row * 2 + col
                ax[row, col].plot(*cell.exterior.xy, lw=5, label="cell")
                ax[row, col].plot(*circle[i].exterior.xy, lw=3, c="g", label="circle")
                ax[row, col].plot(
                    *intersection_shapely[i].exterior.xy,
                    "--",
                    lw=2,
                    c="r",
                    label="intersection"
                )
                ax[row, col].set_xlim(left[0], right[0])
                ax[row, col].set_ylim(bottom[0], top[0])
                ax[row, col].set_ylabel("y (m)")
                ax[row, col].set_xlabel("x (m)")
                ax[row, col].set_aspect("equal")

        plt.tight_layout()
        if self.save_fig:
            plt.savefig(FIGURES_PATH / "test_calculate_case_11_area.png")
        if self.show_fig:
            plt.show()


class TestEncodeCorners:
    def test_case_0(self):
        """No corners inside - should return case 0"""
        top_left = np.array([False])
        top_right = np.array([False])
        bottom_left = np.array([False])
        bottom_right = np.array([False])
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.size == 1
        assert case == 0

    def test_case_1(self):
        """Bottom left corner inside - should return case 1"""
        top_left = np.array([False])
        top_right = np.array([False])
        bottom_left = np.array([True])
        bottom_right = np.array([False])
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.size == 1
        assert case == 1

    def test_case_3(self):
        """bottom left and bottom right corners inside - should return case 3"""
        top_left = np.array([False])
        top_right = np.array([False])
        bottom_left = np.array([True])
        bottom_right = np.array([True])
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.size == 1
        assert case == 3

    def test_case_9(self):
        """top left and bottom left corners inside - should return case 9"""
        top_left = np.array([True])
        top_right = np.array([False])
        bottom_left = np.array([True])
        bottom_right = np.array([False])
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.size == 1
        assert case == 9

    def test_case_11(self):
        """top left, bottom left, and bottom right corners inside - should
        return case 11"""
        top_left = np.array([True])
        top_right = np.array([False])
        bottom_left = np.array([True])
        bottom_right = np.array([True])
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.size == 1
        assert case == 11

    def test_case_15(self):
        """All corners inside - should return case 15"""
        top_left = np.array([True])
        top_right = np.array([True])
        bottom_left = np.array([True])
        bottom_right = np.array([True])
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.size == 1
        assert case == 15

    def test_multiple(self):
        """
        Tests a 3x3 grid of the form:

        3   1 0
        15  11 1
        15 15 9
        """
        top_left = np.array(
            [[False, False, False], [True, True, False], [True, True, True]]
        )
        top_right = np.array(
            [[False, False, False], [True, False, False], [True, True, False]]
        )
        bottom_left = np.array(
            [[True, True, False], [True, True, True], [True, True, True]]
        )
        bottom_right = np.array(
            [[True, False, False], [True, True, False], [True, True, False]]
        )
        case = _encode_corners(top_left, top_right, bottom_left, bottom_right)
        assert isinstance(case, np.ndarray)
        assert case.shape == (3, 3)
        assert case[0, 0] == 3
        assert case[0, 1] == 1
        assert case[0, 2] == 0
        assert case[1, 0] == 15
        assert case[1, 1] == 11
        assert case[1, 2] == 1
        assert case[2, 0] == 15
        assert case[2, 1] == 15
        assert case[2, 2] == 9


class TestComputeIntersectionAreaByCase:
    length = 1

    def test_case_0_no_radius(self):
        """No corners inside - should return 0"""
        case = np.array([[[0]]])
        left = np.array([[[0]]])
        right = np.array([[[0]]])
        bottom = np.array([[[0]]])
        top = np.array([[[0]]])
        radius = np.array([[[0]]])
        intersection_area = _compute_intersection_area_by_case(
            case, self.length, left, right, bottom, top, radius
        )
        assert isinstance(intersection_area, np.ndarray)
        assert intersection_area.size == 1
        assert intersection_area == 0

    def test_case_0_with_radius(self):
        """No corners inside, but with a nonzero radius - should return area
        of the circle"""
        case = np.array([[[0]]])
        left = np.array([[[0]]])
        right = np.array([[[0]]])
        bottom = np.array([[[0]]])
        top = np.array([[[0]]])
        radius = np.array([[[0.25]]])
        intersection_area = _compute_intersection_area_by_case(
            case, self.length, left, right, bottom, top, radius
        )
        assert isinstance(intersection_area, np.ndarray)
        assert intersection_area.size == 1
        assert np.allclose(intersection_area, np.pi * radius[0, 0, 0] ** 2)


class TestSumAreaAlongAxis:
    def test_1D_1_cell(self):
        data = np.array([1, 1])
        summed = _sum_area_along_axis(data, 0, 2)
        assert summed == 2

    def test_1D_2_cells(self):
        data = np.array([1, 1, 1, 1])
        summed = _sum_area_along_axis(data, 0, 2)
        assert np.allclose(summed, [2, 2])

    def test_1D_10_cells(self):
        data = np.ones(10)
        summed = _sum_area_along_axis(data, 0, 2)
        assert np.allclose(summed, np.ones(5) * 2)

    def test_1D_invalid_axis(self):
        data = np.ones(10)
        with pytest.raises(ValueError):
            _sum_area_along_axis(data, 1, 2)

    def test_2D(self):
        data = np.array([[1, 1], [1, 1]])

        # Sum row-wise
        summed = _sum_area_along_axis(data, 0, 2)
        assert summed.shape == (1, 2)
        assert np.allclose(summed, [2, 2])

        # Sum column-wise
        summed = _sum_area_along_axis(data, 1, 2)
        assert summed.shape == (2, 1)
        assert np.allclose(summed, [2, 2])

    def test_2D_two_sums(self):
        data = np.ones([4, 4])

        # Sum row-wise
        summed = _sum_area_along_axis(data, 0, 2)
        assert summed.shape == (2, 4)
        assert np.allclose(summed, np.ones([2, 4]) * 2)

        # Sum column-wise
        summed_again = _sum_area_along_axis(summed, 1, 2)
        assert summed_again.shape == (2, 2)
        assert np.allclose(summed_again, np.ones([2, 2]) * 4)

    def test_3D(self):
        data = np.ones([10, 10, 10])
        summed = _sum_area_along_axis(data, 0, 10)
        assert summed.shape == (1, 10, 10)
        assert np.allclose(summed, np.ones([1, 10, 10]) * 10)

    def test_3D_2_layers(self):
        data = np.ones([20, 10, 10])
        summed = _sum_area_along_axis(data, 0, 10)
        assert summed.shape == (2, 10, 10)
        assert np.allclose(summed, np.ones([2, 10, 10]) * 10)


class TestDiscretizeCrownProfile:
    def test_known_tree_fine_resolution(self):
        test_tree = Tree(122, 1, 24, 20.66443, 0.5)
        grid = discretize_crown_profile(test_tree, 0.5, 0.5)
        assert np.sum(grid) > 1

    def test_known_tree_mod_resolution(self):
        test_tree = Tree(122, 1, 24, 20.66443, 0.5)
        grid = discretize_crown_profile(test_tree, 1, 1)
        assert np.sum(grid) > 1

    def test_known_tree_coarse_resolution(self):
        test_tree = Tree(122, 1, 24, 20.66443, 0.5)
        grid = discretize_crown_profile(test_tree, 2, 1)
        assert np.sum(grid) > 1

    def test_random_trees(self):
        resolutions = ((0.5, 0.5), (1.0, 1.0), (2.0, 1.0))
        for _ in range(1000):
            tree = make_random_tree()
            res_tuple = random.choice(resolutions)
            grid = discretize_crown_profile(tree, *res_tuple)
            assert np.sum(grid) > 0


class TestComputeHorizontalProbability:
    def test_alpha_zero_creates_uniform_distribution(self):
        alpha = 0
        mask = np.ones((10, 10, 10))
        expected = np.ones_like(mask)
        actual = _compute_horizontal_probability(mask, alpha)
        assert np.allclose(actual, expected)

    def test_alpha_zero_with_mask(self):
        alpha = 0
        mask = np.zeros((90, 90, 90))
        mask[:, 20:70, 20:70] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_horizontal_probability(mask, alpha)

        assert np.all(result[outside_mask] == 0)
        assert np.all(result[inside_mask] == 1)

    def test_zero_mask_results_in_zero_probability(self):
        beta = 1
        mask = np.zeros((10, 10, 10))
        actual = _compute_horizontal_probability(mask, beta)
        assert np.all(actual == 0)

    def test_positive_alpha_pushes_probabilities_to_edges(self):
        alpha = 1
        mask = np.zeros((90, 90, 90))
        mask[:, 20:70, 20:70] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_horizontal_probability(mask, alpha)

        # The probability is 0 outside the mask
        assert np.all(result[outside_mask]) == 0

        # The probability is > 0 everywhere inside the mask
        assert np.all(result[inside_mask]) > 0

        # The probability is lowest at the center of the mask and increases
        # as we move away from the center (in the xy plane)
        for i in range(45, 69):
            assert np.all(result[:, i, i] <= result[:, i + 1, i + 1])

        # The min and max probability are 0 and 1
        assert np.min(result) == 0
        assert np.max(result) == 1

    def test_range_positive_alphas(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        for alpha in np.random.random(100):
            result = _compute_horizontal_probability(mask, alpha)

            # The probability is 0 outside the mask
            assert np.all(result[outside_mask]) == 0

            # The probability is > 0 everywhere inside the mask
            assert np.all(result[inside_mask]) > 0

            # The probability is lowest at the center of the mask and increases
            # as we move away from the center (in the xy plane)
            for i in range(4, 6):
                assert np.all(result[:, i, i] <= result[:, i + 1, i + 1])

            # The min and max probability are 0 and 1
            assert np.min(result) == 0
            assert np.max(result) == 1

    def test_negative_alpha_pushes_probabilities_to_center(self):
        alpha = -1
        mask = np.zeros((90, 90, 90))
        mask[:, 20:70, 20:70] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_horizontal_probability(mask, alpha)

        # The probability is 0 outside the mask
        assert np.all(result[outside_mask]) == 0

        # The probability is > 0 everywhere inside the mask
        assert np.all(result[inside_mask]) > 0

        # The probability is highest at the center of the mask and decreases
        # as we move away from the center (in the xy plane)
        for i in range(45, 69):
            assert np.all(result[:, i, i] >= result[:, i + 1, i + 1])

        # The minimum and maximum probabilities are 0 and 1
        assert np.min(result) == 0
        assert np.max(result) == 1

    def test_range_negative_alphas(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        for alpha in -np.random.random(100):
            result = _compute_horizontal_probability(mask, alpha)

            # The probability is 0 outside the mask
            assert np.all(result[outside_mask]) == 0

            # The probability is > 0 everywhere inside the mask
            assert np.all(result[inside_mask]) > 0

            # The probability is highest at the center of the mask and decreases
            # as we move away from the center (in the xy plane)
            for i in range(4, 6):
                assert np.all(result[:, i, i] >= result[:, i + 1, i + 1])

            # The minimum and maximum probabilities are 0 and 1
            assert np.min(result) == 0
            assert np.max(result) == 1

    def test_higher_alpha_pushes_more_probability_to_edges(self):
        alpha_1 = 0.5
        alpha_2 = 1.0
        mask = np.zeros((90, 90, 90))
        mask[:, 20:70, 20:70] = 1
        result_1 = _compute_horizontal_probability(mask, alpha_1)
        result_2 = _compute_horizontal_probability(mask, alpha_2)

        # Assert that result_2 pushes more probability to the edges
        for i in range(45, 69):
            assert np.all(result_1[:, i, i] >= result_2[:, i, i])

        # The maximum probability for both alpha values is 1
        assert np.max(result_1) == 1
        assert np.max(result_2) == 1

    # TODO: TEST MASK VALUES LIKE MASK[20:70, 20:70, 20:70] = 0.5


class TestComputeVerticalProbability:
    def test_beta_zero_creates_uniform_distribution(self):
        beta = 0
        mask = np.ones((10, 10, 10))
        expected = np.ones_like(mask)
        actual = _compute_vertical_probability(mask, beta)
        assert np.allclose(actual, expected)

    def test_beta_zero_with_mask(self):
        alpha = 0
        mask = np.zeros((90, 90, 90))
        mask[:, 20:70, 20:70] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_vertical_probability(mask, alpha)

        assert np.all(result[outside_mask] == 0)
        assert np.all(result[inside_mask] == 1)

    def test_zero_mask_results_in_zero_probability(self):
        beta = 1
        mask = np.zeros((10, 10, 10))
        actual = _compute_vertical_probability(mask, beta)
        assert np.all(actual == 0)

    def test_positive_beta_pushes_probabilities_to_top(self):
        beta = 1
        mask = np.zeros((90, 90, 90))
        mask[:, 20:70, 20:70] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_vertical_probability(mask, beta)

        # The probability is 0 outside the mask
        assert np.all(result[outside_mask]) == 0

        # The probability is > 0 everywhere inside the mask
        assert np.all(result[inside_mask]) > 0

        # The probability is lowest at the bottom of the mask and increases
        # as we move towards the top (along the z-axis)
        for i in range(89):
            assert np.all(result[i, ...] <= result[i + 1, ...])

        # The minimum and maximum probabilities are 0 and 1
        assert np.min(result) == 0
        assert np.max(result) == 1

    def test_range_positive_betas(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        for beta in np.random.random(100):
            result = _compute_vertical_probability(mask, beta)
            # The probability is 0 outside the mask
            assert np.all(result[outside_mask]) == 0

            # The probability is > 0 everywhere inside the mask
            assert np.all(result[inside_mask]) > 0

            # The probability is lowest at the bottom of the mask and increases
            # as we move towards the top (along the z-axis)
            for i in range(8):
                assert np.all(result[i, :, :] <= result[i + 1, :, :])

            # The minimum and maximum probabilities are 0 and 1
            assert np.min(result) == 0
            assert np.max(result) == 1

    def test_higher_beta_values_increase_exponentially(self):
        beta_1 = 1
        beta_2 = 2
        mask = np.ones((10, 10, 10))
        actual_1 = _compute_vertical_probability(mask, beta_1)
        actual_2 = _compute_vertical_probability(mask, beta_2)

        # Assert that actual_2 is less than actual_1 for all z, except at the
        # top of the mask, where both are 1
        for z in range(9):
            assert np.all(actual_1[z, :, :] >= actual_2[z, :, :])

        # The maximum probability for both beta values is 1
        assert np.max(actual_1) == 1
        assert np.max(actual_2) == 1


class TestComputeJointProbability:

    def test_uniform_joint_probability(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_joint_probability(mask=mask, alpha=0, beta=0)

        # The probability is 0 outside the mask
        assert np.all(result[outside_mask] == 0)

        # The probability is > 0 everywhere inside the mask
        assert np.all(result[inside_mask] > 0)

        # The minimum and maximum probabilities are 0 and 1
        assert np.min(result) == 0
        assert np.max(result) == 1

        # Probability is the same everywhere inside the mask
        assert np.all(result[inside_mask] == result[inside_mask][0])

    def test_positive_alpha_zero_beta(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_joint_probability(mask=mask, alpha=1, beta=0)

        # The probability is 0 outside the mask
        assert np.all(result[outside_mask] == 0)

        # The probability is > 0 everywhere inside the mask
        assert np.all(result[inside_mask] > 0)

        # Minimum and maximum probabilities are 0 and 1
        assert np.min(result) == 0
        assert np.max(result) == 1

        # The horizontal probability (xy plane) is lowest at the edges of the
        # mask and increases as we move towards the center
        for i in range(4, 6):
            assert np.all(result[:, i, i] <= result[:, i + 1, i + 1])

        # The vertical probability is uniform along the z-axis
        for i in range(8):
            assert np.all(result[i, ...] == result[i + 1, ...])

    def test_zero_alpha_positive_beta(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        result = _compute_joint_probability(mask=mask, alpha=0, beta=1)

        # The probability is 0 outside the mask
        assert np.all(result[outside_mask] == 0)

        # The probability is > 0 everywhere inside the mask
        assert np.all(result[inside_mask] > 0)

        # Minimum and maximum probabilities are 0 and 1
        assert np.min(result) == 0
        assert np.max(result) == 1

        # The horizontal probability (xy plane) is uniform along the x/y axes
        for i in range(2, 6):
            assert np.all(result[:, i, i] == result[:, i + 1, i + 1])

        # The vertical probability is lowest at the bottom of the mask and
        # increases as we move towards the top (along the z-axis)
        for i in range(8):
            assert np.all(result[i, ...] <= result[i + 1, ...])

    def test_positive_alpha_positive_beta(self):
        mask = np.zeros((9, 9, 9))
        mask[:, 2:7, 2:7] = 1
        inside_mask = np.where(mask == 1)
        outside_mask = np.where(mask == 0)
        for alpha in np.random.random(10):
            for beta in np.random.random(10):
                result = _compute_joint_probability(mask=mask, alpha=alpha, beta=beta)

                # The probability is 0 outside the mask
                assert np.all(result[outside_mask] == 0)

                # The probability is > 0 everywhere inside the mask
                assert np.all(result[inside_mask] > 0)

                # Minimum and maximum probabilities are 0 and 1
                assert np.min(result) == 0
                assert np.max(result) == 1

                # The horizontal probability (xy plane) is lowest at the edges
                # of the mask and increases as we move towards the center
                for i in range(4, 6):
                    assert np.all(result[:, i, i] <= result[:, i + 1, i + 1])

                # The vertical probability is lowest at the bottom of the mask
                # and increases as we move towards the top (along the z-axis)
                for i in range(8):
                    assert np.all(result[i, ...] <= result[i + 1, ...])

    def test_multiple_masks_equal(self):
        for j in range(10):
            alpha = np.random.random()
            beta = np.random.random()
            results_list = []
            for i in range(5):
                mask = np.zeros((9, 9, 9))
                mask[:, 2:7, 2:7] = 1
                result = _compute_joint_probability(mask=mask, alpha=alpha, beta=beta)
                results_list.append(result)

            # Assert that all results are equal
            for i in range(1, len(results_list)):
                assert np.all(results_list[i] == results_list[i - 1])


class TestSampleCrownVoxels:

    def test_basic_functionality(self):
        """Test basic functionality"""
        joint_probability = np.ones((3, 3, 3))
        n = 5
        result = _sample_voxels_from_probability_grid(n, joint_probability, 42)
        assert np.sum(result) > 0
        assert np.count_nonzero(result) == n

    def test_all_zero_probabilities(self):
        """Test with all zero probabilities"""
        joint_probability = np.zeros((3, 3, 3))
        n = 5
        result = _sample_voxels_from_probability_grid(n, joint_probability, 42)
        assert np.sum(result) == 0
        assert np.count_nonzero(result) == 0

    def test_n_greater_than_voxels(self):
        """Test with n greater than the number of voxels"""
        joint_probability = np.ones((3, 3, 3))
        n = joint_probability.size + 1
        with pytest.raises(ValueError):
            _sample_voxels_from_probability_grid(n, joint_probability, 42)

    def test_negative_n(self):
        """Test with negative n"""
        joint_probability = np.ones((3, 3, 3))
        n = -5
        with pytest.raises(ValueError):
            _sample_voxels_from_probability_grid(n, joint_probability, 42)

    def test_n_zero(self):
        """Test with n equals zero"""
        joint_probability = np.ones((3, 3, 3))
        n = 0
        result = _sample_voxels_from_probability_grid(n, joint_probability, 42)
        assert np.sum(result) == 0
        assert np.count_nonzero(result) == 0

    def test_n_range(self):
        """Tests a range of values for n from 1 to the number of voxels"""
        joint_probability = np.ones((3, 3, 3))
        for n in range(1, joint_probability.size + 1):
            result = _sample_voxels_from_probability_grid(n, joint_probability, 42)
            assert np.sum(result) > 0
            assert np.count_nonzero(result) == n

    def test_non_uniform_probabilities(self):
        """Test with non-uniform probabilities"""
        joint_probability = np.array(
            [
                [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
                [[0.9, 0.8, 0.7], [0.6, 0.5, 0.4], [0.3, 0.2, 0.1]],
                [[0.1, 0.2, 0.3], [0.4, 0.5, 0.6], [0.7, 0.8, 0.9]],
            ]
        )
        n = joint_probability.size // 2
        result = _sample_voxels_from_probability_grid(n, joint_probability, 42)
        assert np.sum(result) > 0
        assert np.count_nonzero(result) == n

        # Check that cells with higher probabilities are more likely to be selected
        selected = np.where(result > 0, 1, 0)
        selected_cells_prob = np.sum(joint_probability[selected == 1])
        unselected_cells_prob = np.sum(joint_probability[selected == 0])
        assert selected_cells_prob > unselected_cells_prob

    def test_same_seed(self):
        """Test if the random seed works as expected"""
        joint_probability = np.ones((3, 3, 3))
        n = 5

        result1 = _sample_voxels_from_probability_grid(n, joint_probability, 42)
        result2 = _sample_voxels_from_probability_grid(n, joint_probability, 42)

        assert np.array_equal(result1, result2)

    def test_different_seed(self):
        """Test if the random seed works as expected"""
        joint_probability = np.ones((3, 3, 3))
        n = 5

        result1 = _sample_voxels_from_probability_grid(n, joint_probability, 42)
        result2 = _sample_voxels_from_probability_grid(n, joint_probability, 43)

        assert not np.array_equal(result1, result2)


class TestVoxelizedTree:
    def test_distribute_biomass(self):
        for _ in range(1000):
            tree = make_random_tree(status_code=1)
            voxelized_tree = voxelize_tree(tree, 1, 1)
            biomass_array = voxelized_tree.distribute_biomass()
            distributed_biomass = np.sum(biomass_array)
            assert np.allclose(distributed_biomass, tree.foliage_biomass)



# Core imports
from __future__ import annotations
import warnings
from enum import Enum
from typing import Protocol

# External imports
import numpy as np
from pandas import DataFrame


class ThinningDirection(str, Enum):
    """
    Enumeration of thinning directions. Thinning from below is the same as low
    thinning and thinning from above is the same as crown thinning.

    Attributes
    ----------
    BELOW : str
        Thinning from below (low thinning).
    ABOVE : str
        Thinning from above (crown thinning).
    """

    BELOW: str = "below"
    ABOVE: str = "above"


class TreatmentProtocol(Protocol):
    """
    Protocol for all treatment implementations.

    Methods
    -------
    apply(trees: DataFrame) -> DataFrame
        Apply the treatment to the DataFrame of trees.
    """

    def apply(self, trees: DataFrame) -> DataFrame:
        pass


class DirectionalThinToDiameterLimit:
    """
    Thinning treatment to limit trees based on their diameter.

    Parameters
    ----------
    limit : float
        Diameter limit for thinning, in centimeters (cm).
    direction : ThinningDirection
        Direction of thinning, either 'below' or 'above'. By default, 'below'.

    Methods
    -------
    apply(trees: DataFrame, dia_column_name: str = "DIA") -> DataFrame
        Apply the diameter limit thinning to the DataFrame of trees.
    """

    def __init__(
        self, limit: float, direction: ThinningDirection = ThinningDirection.BELOW
    ) -> None:
        self.limit = limit
        self.direction = direction

    def apply(self, trees: DataFrame, dia_column_name: str = "DIA") -> DataFrame:
        """
        Apply the diameter limit thinning to the DataFrame of trees.

        Parameters
        ----------
        trees : DataFrame
            DataFrame containing tree data.
        dia_column_name : str, optional
            Name of the diameter column in the DataFrame, by default "DIA".
            The diameter values should be in centimeters (cm).

        Returns
        -------
        DataFrame
            DataFrame after applying the diameter limit thinning.

        Raises
        ------
        ValueError
            If the thinning direction is invalid.
        """
        df = trees.copy()

        if self.direction == ThinningDirection.BELOW:
            df = df[df[dia_column_name] >= self.limit]
        elif self.direction == ThinningDirection.ABOVE:
            df = df[df[dia_column_name] < self.limit]
        else:
            raise ValueError("Invalid thinning direction. Use 'below' or 'above'.")

        assert isinstance(df, DataFrame), "Resulting object is not a DataFrame"

        return df


class DirectionalThinToStandBasalArea:
    """
    Thinning treatment to limit the stand basal area.

    Parameters
    ----------
    target : float
        Target basal area for thinning, in square meters (m²).
    direction : ThinningDirection
        Direction of thinning, either 'below' or 'above'. By default, 'below'.

    Methods
    -------
    apply(trees: DataFrame, dia_column_name: str = "DIA") -> DataFrame
        Apply the basal area limit thinning to the DataFrame of trees.
    """

    def __init__(
        self, target: float, direction: ThinningDirection = ThinningDirection.BELOW
    ) -> None:
        self.target = target
        self.direction = direction

    def apply(self, trees: DataFrame, dia_column_name: str = "DIA") -> DataFrame:
        """
        Apply the basal area limit thinning to the DataFrame of trees.

        Parameters
        ----------
        trees : DataFrame
            DataFrame containing tree data.
        dia_column_name : str, optional
            Name of the diameter column in the DataFrame, by default "DIA".
            The diameter values should be in centimeters (cm).

        Returns
        -------
        DataFrame
            DataFrame after applying the basal area limit thinning.

        Raises
        ------
        ValueError
            If the thinning direction is invalid.
        """
        df = trees.copy()

        # Calculate basal area for each tree in square meters
        df["BA"] = df[dia_column_name] ** 2 * (np.pi / 40_000)

        if df["BA"].sum() <= self.target:
            return df

        if self.direction == ThinningDirection.BELOW:
            df.sort_values(by=dia_column_name, ascending=False, inplace=True)
        elif self.direction == ThinningDirection.ABOVE:
            df.sort_values(by=dia_column_name, ascending=True, inplace=True)
        else:
            raise ValueError("Invalid thinning direction. Use 'below' or 'above'.")

        df["BA_CUMSUM"] = df["BA"].cumsum()
        df = df[df["BA_CUMSUM"] <= self.target]

        assert isinstance(df, DataFrame), "Resulting object is not a DataFrame"

        return df.drop(columns=["BA", "BA_CUMSUM"])


class ProportionalThinToBasalArea:
    """
    Proportional thinning treatment to reach a target basal area.

    Parameters
    ----------
    target : float
        Target basal area for thinning, in square meters (m²).

    Methods
    -------
    apply(trees: DataFrame, dia_column_name: str = "DIA") -> DataFrame
        Apply the proportional thinning to reach the target basal area.
    """

    def __init__(self, target: float) -> None:
        self.target = target

    def apply(self, trees: DataFrame, dia_column_name: str = "DIA") -> DataFrame:
        """
        Apply the proportional thinning to reach the target basal area.

        Parameters
        ----------
        trees : DataFrame
            DataFrame containing tree data.
        dia_column_name : str, optional
            Name of the diameter column in the DataFrame, by default "DIA".
            The diameter values should be in centimeters (cm).

        Returns
        -------
        DataFrame
            DataFrame after applying the proportional thinning to the target basal area.

        Warns
        -----
        RuntimeWarning
            If the resulting basal area is still above the target after thinning.
        """
        df = trees.copy()

        # Calculate basal area for each tree in square meters
        df["BA"] = df[dia_column_name] ** 2 * (np.pi / 40_000)

        total_basal_area = df["BA"].sum()

        if total_basal_area <= self.target:
            return df

        proportion_to_remove = 1 - (self.target / total_basal_area)
        df["remove"] = np.random.rand(len(df)) < proportion_to_remove
        df = df[~df["remove"]]

        result_ba = df["BA"].sum()
        if result_ba > self.target:
            warnings.warn(
                f"Resulting basal area ({result_ba:.4f} m²) is still above the target ({self.target:.4f} m²) "
                f"after proportional thinning. Difference: {result_ba - self.target:.4f} m²",
                RuntimeWarning,
            )

        assert isinstance(df, DataFrame), "Resulting object is not a DataFrame"

        return df.drop(columns=["BA", "remove"])



# Core imports
from __future__ import annotations
from typing import TYPE_CHECKING

# Internal imports
if TYPE_CHECKING:
    from fastfuels_core.trees import Tree

# External imports
import numpy as np
from numpy import ndarray
from scipy.ndimage import distance_transform_edt


class VoxelizedTree:
    def __init__(self, tree: "Tree", grid: ndarray, hr, vr):
        self.tree = tree
        self.grid = grid
        self.hr = hr
        self.vr = vr

    def distribute_biomass(self):
        volume = np.sum(self.grid) * self.hr * self.hr * self.vr
        foliage_mpv = self.tree.foliage_biomass / volume
        biomass_grid = self.grid.copy() * foliage_mpv
        return biomass_grid


def voxelize_tree(
    tree: "Tree",
    horizontal_resolution: float,
    vertical_resolution: float,
    **kwargs,
) -> VoxelizedTree:

    crown_profile_mask = discretize_crown_profile(
        tree, horizontal_resolution, vertical_resolution
    )

    alpha = kwargs.get("alpha", 0.5)
    beta = kwargs.get("beta", 0.5)
    rho = kwargs.get("rho", None)
    seed = kwargs.get("seed", None)

    sampled_crown_mask = sample_occupied_cells(
        crown_profile_mask, alpha, beta, rho, seed
    )
    return VoxelizedTree(
        tree, sampled_crown_mask, horizontal_resolution, vertical_resolution
    )


def discretize_crown_profile(
    tree: "Tree", hr: float, vr: float, full_intersection=True
) -> ndarray:
    # Get the horizontal and vertical coordinates of the tree crown
    horizontal_coords = _get_horizontal_tree_coords(hr, tree.max_crown_radius)
    z_pts = _get_vertical_tree_coords(vr, tree.height, tree.crown_base_height)

    # Slice the horizontal coordinates to get the first quadrant of the xy plane
    q2_slice = slice(len(horizontal_coords) // 2, None)
    x_pts_q2 = horizontal_coords[q2_slice]
    y_pts_q2 = np.flip(x_pts_q2)

    q2_grid = _discretize_crown_profile_quadrant(
        tree, x_pts_q2, y_pts_q2, z_pts, hr, vr, full_intersection
    )

    # Build the other quadrants by flipping the q2 grid about the x and y axes
    # Note that q2 grid has dimensions (z, y, x)
    q1_grid = np.flip(q2_grid, axis=2)
    q3_grid = np.flip(q2_grid, axis=1)
    q4_grid = np.flip(q3_grid, axis=2)

    return _align_quadrants(q1_grid, q2_grid, q3_grid, q4_grid)


def _get_vertical_tree_coords(step, tree_height, crown_base_height):
    """
    Returns a grid of coordinates for a tree of height, height, with a spacing
    step. The grid is returned as a 1D array.
    """
    grid = np.arange(crown_base_height, tree_height + step, step)
    return grid


def _get_horizontal_tree_coords(step, radius, pos=0.0):
    """
    Discretizes a stem position and crown radius into a 1D array of coordinates
    centered at pos, with a spacing step. The grid has an odd number of cells.
    The grid is rounded out to include one cell beyond the crown radius.
    """
    cells_per_side = int(np.floor(np.abs(radius / step))) + 1
    lower_bound = pos - cells_per_side * step
    upper_bound = pos + cells_per_side * step
    grid = np.linspace(lower_bound, upper_bound, 2 * cells_per_side + 1)
    return grid


def _discretize_crown_profile_quadrant(
    tree: "Tree", x_pts, y_pts, z_pts, hr, vr, full_intersection=False
):
    """
    Build a 3D grid of a quadrant of a tree crown represented as a rotational
    solid.
    """
    # Create a subgrid of the z-axis to increase the resolution of the crown
    vr_subgrid = 0.1
    num_subgrid_cells_per_z = int(vr / vr_subgrid)
    z_pts_subgrid = _resample_coords_grid_to_subgrid(z_pts, vr, vr_subgrid)
    r_at_height_z = tree.get_crown_radius_at_height(z_pts_subgrid)

    # Create a 3D grid of the tree crown and compute the area of intersection
    # between the tree crown and each cell of the grid along the z-axis
    r_grid, y_grid, x_grid = np.meshgrid(r_at_height_z, y_pts, x_pts, indexing="ij")
    area = _compute_intersection_area(x_grid, y_grid, hr, r_grid, full_intersection)

    # Convert the area of intersection to a volume fraction by summing the area
    # along the z-axis and dividing by the cell volume
    volume_subgrid = area * vr_subgrid
    volume = _sum_area_along_axis(volume_subgrid, 0, num_subgrid_cells_per_z)
    volume_fraction = np.minimum(volume / (hr * hr * vr), 1.0)

    return volume_fraction


def _resample_coords_grid_to_subgrid(
    grid: ndarray, grid_spacing: float, new_spacing: float
) -> ndarray:
    """
    Resamples grid with spacing grid_spacing to a subgrid with spacing
    new_spacing.
    """
    subgrid = np.arange(
        grid[0] - grid_spacing / 2 + new_spacing / 2,
        grid[-1] + grid_spacing / 2,
        new_spacing,
    )

    return subgrid


def _compute_intersection_area(
    x_center: ndarray, y_center: ndarray, length: float, radius: ndarray, exact=False
):
    # Calculate location of cell edges using the cell center and length
    x_right = x_center + length / 2
    x_left = x_center - length / 2
    y_top = y_center + length / 2
    y_bottom = y_center - length / 2

    # Calculate the distance from the origin to each cell's corner points
    top_left_dist = np.sqrt(x_left**2 + y_top**2)
    top_right_dist = np.sqrt(x_right**2 + y_top**2)
    bottom_left_dist = np.sqrt(x_left**2 + y_bottom**2)
    bottom_right_dist = np.sqrt(x_right**2 + y_bottom**2)

    # Check if each corner of the cell is inside the circle
    top_left_in = top_left_dist < radius
    top_right_in = top_right_dist < radius
    bottom_left_in = bottom_left_dist < radius
    bottom_right_in = bottom_right_dist < radius

    case_index = _encode_corners(
        top_left_in, top_right_in, bottom_left_in, bottom_right_in
    )

    area = _compute_intersection_area_by_case(
        case_index, length, x_left, x_right, y_bottom, y_top, radius, exact
    )

    return area


def _encode_corners(
    top_left_inside, top_right_inside, bottom_left_inside, bottom_right_inside
):
    """
    Encode the inside/outside status of the corners of a cell into an index.
    Each array of inside/outside status is a boolean array of the same shape
    as the cell grid. The index is computed as the sum of the inside/outside
    status of each corner in binary representation.
    """
    case_index = (
        top_left_inside * 8
        + top_right_inside * 4
        + bottom_right_inside * 2
        + bottom_left_inside
    )
    return case_index


def _compute_intersection_area_by_case(
    case_index, length, left, right, bottom, top, radius, exact=False
):
    # Initialize an array to hold the computed areas
    areas = np.zeros_like(case_index, dtype=float)

    # Case 0a: No corners inside, no intersection
    areas[case_index == 0] = 0.0

    # Case 0b: No corners inside, but circle is inside cell
    areas[:, -1, 0] = np.pi * radius[:, -1, 0] ** 2

    # Case 1: Bottom left inside
    case_1 = case_index == 1
    areas[case_1] = _calculate_case_1_area(
        left[case_1], bottom[case_1], radius[case_1], exact
    )

    # Case 3: Bottom left and bottom right inside
    case_3 = case_index == 3
    areas[case_3] = _calculate_case_3_area(
        left[case_3],
        right[case_3],
        bottom[case_3],
        radius[case_3],
        length,
        exact,
    )

    # Case 9: Top left and bottom left inside
    case_9 = case_index == 9
    areas[case_9] = _calculate_case_9_area(
        top[case_9],
        bottom[case_9],
        left[case_9],
        radius[case_9],
        length,
        exact,
    )

    # Case 11: Top left, bottom left, and bottom right inside
    case_11 = case_index == 11
    areas[case_11] = _calculate_case_11_area(
        top[case_11], right[case_11], radius[case_11], length, exact
    )

    # Case 15: All corners inside, full cell area
    areas[case_index == 15] = np.square(length)

    return areas


def _calculate_case_1_area(left_edge, bottom_edge, radius, exact=False):
    """
    Calculate the area of intersection between a circle and a cell when the
    bottom left corner of the cell is inside the circle.

    Area is approximated as a triangle whose height is the distance from the
    bottom left corner of the cell to the circle's intersection with the
    left edge of the cell, and whose base is the distance from the bottom
    left corner of the cell to the circle's intersection with the bottom
    edge of the cell.

    For an exact area calculation, the area of the circular segment is added to
    the area of the triangle.
    """
    p_x = left_edge
    p_y = _find_circle_cell_intersection_coord(radius, p_x)
    q_y = bottom_edge
    q_x = _find_circle_cell_intersection_coord(radius, q_y)

    triangle_area = _compute_triangle_area(p_y - bottom_edge, q_x - left_edge)
    if exact:
        circular_segment_area = _compute_circle_segment_area(p_x, p_y, q_x, q_y, radius)
        return triangle_area + circular_segment_area

    return triangle_area


def _calculate_case_3_area(
    left_edge, right_edge, bottom_edge, radius, length, exact=False
):
    """
    Calculate the area of intersection between a circle and a cell when the
    bottom left and bottom right corners of the cell are inside the circle.

    Area is approximated by the areas of a triangle and a rectangle. The
    triangle is given by a polygon whose height is the distance from the
    height of the circle's intersection with the cell on the right edge to
    the height of the circle's intersection with the cell on the left edge,
    and whose base is the width of the cell. The rectangle's height is given by
    the distance from the bottom right corner to the circle's intersection with
    the right edge, and its width is the width of the cell.

    For an exact area calculation, the area of the circular segment is added to
    the area of the triangle and rectangle.
    """
    p_x = left_edge
    p_y = _find_circle_cell_intersection_coord(radius, p_x)
    q_x = right_edge
    q_y = _find_circle_cell_intersection_coord(radius, q_x)

    triangle_area = _compute_triangle_area(p_y - q_y, length)
    rectangle_area = _compute_rectangle_area(q_y - bottom_edge, length)
    if exact:
        circular_segment_area = _compute_circle_segment_area(p_x, p_y, q_x, q_y, radius)
        return triangle_area + rectangle_area + circular_segment_area

    return triangle_area + rectangle_area


def _calculate_case_9_area(
    top_edge, bottom_edge, left_edge, radius, length, exact=False
):
    """
    Calculate the area of intersection between a circle and a cell when the
    top left and bottom left corners of the cell are inside the circle.

    Area is approximated by the areas of a triangle and a rectangle. The
    triangle is given by a polygon whose height is the height of the cell,
    and whose base is the width from the circle's intersection with the top
    edge to the circle's intersection with the bottom edge. The rectangle's
    height is given by the height of the cell, and its width is the distance
    from the bottom left corner to the circle's intersection with the bottom
    edge.

    For an exact area calculation, the area of the circular segment is added to
    the area of the triangle and rectangle.
    """
    p_y = top_edge
    p_x = _find_circle_cell_intersection_coord(radius, p_y)
    q_y = bottom_edge
    q_x = _find_circle_cell_intersection_coord(radius, q_y)

    triangle_area = _compute_triangle_area(q_x - p_x, length)
    rectangle_area = _compute_rectangle_area(p_x - left_edge, length)
    if exact:
        circular_segment_area = _compute_circle_segment_area(p_x, p_y, q_x, q_y, radius)
        return triangle_area + circular_segment_area + rectangle_area

    return triangle_area + rectangle_area


def _calculate_case_11_area(top_edge, right_edge, radius, length, exact=False):
    """
    Calculate the area of intersection between a circle and a cell when the
    top left, bottom left, and bottom right corners of the cell are inside
    the circle.

    Area is approximated by the area of a trapezoid. The area of the trapezoid
    is given by the area of the cell, minus the area of the triangle formed by
    the circle's intersection with the top edge and the circle's intersection
    with the right edge.

    For an exact area calculation, the area of the circular segment is
    subtracted from the area of the triangle.
    """
    p_y = top_edge
    p_x = _find_circle_cell_intersection_coord(radius, p_y)
    q_x = right_edge
    q_y = _find_circle_cell_intersection_coord(radius, q_x)

    triangle_area = _compute_triangle_area(right_edge - p_x, top_edge - q_y)
    if exact:
        circular_segment_area = _compute_circle_segment_area(p_x, p_y, q_x, q_y, radius)
        return length**2 - (triangle_area - circular_segment_area)

    return length**2 - triangle_area


def _find_circle_cell_intersection_coord(radius, known_coord):
    return np.sqrt(radius**2 - known_coord**2)


def _compute_circle_segment_area(p_x, p_y, q_x, q_y, radius):
    # Compute chord length as the distance between the two intersection points
    chord_length = np.sqrt((p_x - q_x) ** 2 + (p_y - q_y) ** 2)

    # Compute the angle subtended by the circular segment
    theta = _compute_central_angle(chord_length, radius)

    # Compute the area of the circular segment
    circular_segment_area = 0.5 * radius**2 * (theta - np.sin(theta))

    return np.nan_to_num(circular_segment_area)


def _compute_central_angle(chord_length, radius):
    return 2 * np.arcsin(chord_length / (2 * radius))


def _compute_triangle_area(base, height):
    return 0.5 * base * height


def _compute_rectangle_area(base, height):
    return base * height


def _sum_area_along_axis(area: ndarray, axis: int, cells_per_axis: int) -> ndarray:
    """Sum the area along a specified axis to the desired grid resolution."""
    try:
        return np.add.reduceat(
            area, np.arange(0, area.shape[axis], cells_per_axis), axis=axis
        )
    except IndexError:
        raise ValueError("Invalid axis index.")


def _align_quadrants(q1, q2, q3, q4):
    """
    Align four quadrants into a single grid. This function assumes that the
    four quadrants share the same origin and that the quadrants are aligned
    along the x and y axes.
    """
    # Create a grid to hold the four quadrants
    num_x = q1.shape[2] + q2.shape[2] - 1
    num_y = q1.shape[1] + q3.shape[1] - 1
    num_z = q1.shape[0]
    grid = np.zeros((num_z, num_y, num_x))

    mid_x = num_x // 2
    mid_y = num_y // 2

    grid[:, : mid_y + 1, : mid_x + 1] = q1
    grid[:, : mid_y + 1, mid_x:] = q2
    grid[:, mid_y:, mid_x:] = q3
    grid[:, mid_y:, : mid_x + 1] = q4

    return grid


def sample_occupied_cells(
    volume_fraction_array: ndarray,
    alpha: float,
    beta: float,
    rho: float = None,
    seed: int = None,
) -> ndarray:
    # Create a probability mask for the crown grid
    mask_bool = np.where(volume_fraction_array > 0.0, 1.0, 0.0)
    mask_prob = _compute_joint_probability(mask_bool, alpha, beta)

    # Choose n voxels from the joint probability grid
    if rho is None:
        rho = _estimate_crown_density(np.sum(mask_bool))
    n = int(np.count_nonzero(mask_bool) * rho)
    sampled = _sample_voxels_from_probability_grid(n, mask_prob, seed)

    # Make non-zero selected voxels 1
    selected = np.where(sampled > 0, 1.0, 0.0)

    return selected * volume_fraction_array


def _compute_joint_probability(mask: ndarray, alpha: float, beta: float) -> ndarray:
    """
    Combines the horizontal and vertical probability spatial to create a joint
    probability grid for the crown mask.
    """
    # Build the horizontal and vertical probability spatial
    horizontal_probability = _compute_horizontal_probability(mask, alpha)
    vertical_probability = _compute_vertical_probability(mask, beta)

    # Combine the probability spatial for joint probability
    joint_probability = horizontal_probability * vertical_probability
    joint_probability /= np.max(joint_probability)

    return joint_probability


def _compute_horizontal_probability(mask: ndarray, alpha: float) -> ndarray:
    """
    Builds a horizontal probability grid from a binary mask using a Euclidean
    Distance Transform (EDT). The function computes the EDT of the input
    mask, inverts the EDT values by subtracting them from the maximum value,
    applies a power function using the alpha parameter, and finally
    normalizes the result. Any NaN values in the final probability grid are
    replaced with 0.
    """
    # Compute the Euclidean Distance Transform (EDT) of the mask
    edt = distance_transform_edt(mask)

    # Inverse the distance transform
    # Add a small value to avoid zero probabilities
    max_dist = np.max(edt) + 1e-6
    edt = max_dist - edt
    edt[edt == max_dist] = 0

    # Apply the alpha parameter
    edt = np.nan_to_num(mask * edt**alpha)

    # Normalize and replace nans with 0
    horizontal_probability = edt / np.max(edt)
    horizontal_probability = np.nan_to_num(horizontal_probability)

    return horizontal_probability


def _compute_vertical_probability(mask: ndarray, beta: float) -> ndarray:
    """
    Builds a vertical probability grid from a binary mask. The function
    computes a 1D grid along the vertical axis (axis=2) of the input mask,
    raises this grid to the power of beta, and then broadcasts this
    transformed grid across the 2D plane of each layer of the mask (along
    axis=0 and axis=1), and multiplies it with the mask. Finally, the result
    is normalized and any NaN values are replaced with 0.
    """
    # Create a grid for the vertical axis (axis=2)
    z_grid = np.arange(mask.shape[0]).astype(float)

    # Compute the power of the vertical axis grid with beta
    z_power_beta = z_grid**beta

    # Add a small value to avoid zero probabilities
    z_power_beta += 0.01

    # Broadcast the vertical probability grid along axis=0 and axis=1
    vertical_probability = mask * z_power_beta[:, np.newaxis, np.newaxis]

    # Normalize and replace nans with 0
    vertical_probability /= np.max(vertical_probability)
    vertical_probability = np.nan_to_num(vertical_probability)

    return vertical_probability


def _estimate_crown_density(volume: float, threshold: float = 16) -> float:
    if volume < threshold:
        return 1
    return 0.5


def _sample_voxels_from_probability_grid(
    n: int, joint_probability: ndarray, seed: int = None
) -> ndarray:
    """
    Samples n voxels from the joint probability grid. Sampling is weighted by
    the joint probability of each voxel, such that voxels with higher joint
    probability are more likely to be sampled. Voxels are sampled without
    replacement.
    """
    if seed:
        np.random.seed(seed)

    # If joint probability is all zeros, return an empty array
    if np.all(joint_probability == 0):
        return joint_probability

    # Flatten and normalize the joint probability to sum to one
    jp_flat = joint_probability.flatten()
    jp_flat = jp_flat / np.sum(jp_flat)

    # If jp_flat contains nan values, return an empty array
    if np.any(np.isnan(jp_flat)):
        return joint_probability

    # Choose n indices from the flattened joint probability
    chosen_indices = np.random.choice(
        np.arange(joint_probability.size), n, replace=False, p=jp_flat
    )

    # Build flat selection array and reshape to original shape
    selected_flat = np.zeros(joint_probability.size)
    selected_flat[chosen_indices] = jp_flat[chosen_indices]
    return selected_flat.reshape(joint_probability.shape)



# Core imports
from __future__ import annotations
import json
from abc import ABC, abstractmethod
from importlib.resources import files

# Internal imports
from fastfuels_core.base import ObjectIterableDataFrame
from fastfuels_core.point_process import run_point_process
from fastfuels_core.voxelization import VoxelizedTree, voxelize_tree
from fastfuels_core.treatments import TreatmentProtocol

# External Imports
import numpy as np
from numpy import ndarray
from scipy.special import beta
from nsvb.estimators import total_foliage_dry_weight
from pandera import DataFrameSchema, Column, Check, Index


DATA_PATH = files("fastfuels_core.data")
with open(DATA_PATH / "spgrp_parameters.json", "r") as f:
    SPGRP_PARAMS = json.load(f)
with open(DATA_PATH / "spcd_parameters.json", "r") as f:
    SPCD_PARAMS = json.load(f)
with open(DATA_PATH / "class_parameters.json", "r") as f:
    CLASS_PARAMS = json.load(f)


TREE_SCHEMA_COLS = {
    "TREE_ID": Column(int),
    "SPCD": Column(
        int, title="Species Code", description="An FIA integer species code"
    ),
    "STATUSCD": Column(
        int,
        checks=Check.isin([0, 1, 2, 3]),
        title="Status Code",
        description="0 = No status, 1 = Live, 2 = Dead, 3 = Missing",
    ),
    "DIA": Column(
        float,
        checks=[
            Check.gt(0),
            Check.lt(1200),
        ],
        nullable=True,
        title="Diameter at breast height (cm)",
        description="Diameter of the tree measured at breast height (1.37 m)",
    ),
    "HT": Column(
        float,
        checks=[
            Check.gt(0),
            Check.le(116),
        ],
        nullable=True,
        title="Height (m)",
        description="Height of the tree measured from the ground",
    ),
    "CR": Column(
        float,
        checks=Check.in_range(min_value=0, max_value=1),
        nullable=True,
        title="Crown Ratio",
        description="Ratio of the crown length to the total tree height",
    ),
}


class TreeSample(ObjectIterableDataFrame):
    schema = DataFrameSchema(
        columns={
            **TREE_SCHEMA_COLS,
            **{
                "PLOT_ID": Column(int, required=False),
                "TPA": Column(
                    dtype=float,
                    nullable=True,
                    title="Trees per Area (1/m^2)",
                    description="The number of trees per unit area that the sample "
                    "tree represents based on the plot design",
                ),
            },
        },
        coerce=True,
        index=Index(int, unique=True),
    )

    def expand_to_roi(self, process, roi, **kwargs):
        """
        Expands the trees to the Region of Interest (ROI) using a specified
        point process.

        This method generates a new set of trees that fit within the ROI
        based on the characteristics of the specified point process. It
        utilizes the 'run_point_process' function from the 'point_process'
        module to create an instance of the PointProcess class, generate tree
        locations using the specified point process, and assigns these
        locations to trees.

        Parameters
        ----------
        process : str
            The type of point process to use for expanding the trees.
            Currently, only 'inhomogeneous_poisson' is supported.
        roi : GeoDataFrame
            The Region of Interest to which the trees should be expanded. It
            should be a GeoDataFrame containing the spatial boundaries of the
            region.
        **kwargs : dict
            Additional parameters to pass to the point process. See the
            documentation for the 'point_process' module for more information.

        Returns
        -------
        Trees
            A new Trees instance where the trees have been expanded to fit
            within the ROI based on the specified point process.

        Example
        -------
        # TODO: Add example
        """
        return TreePopulation(run_point_process(process, roi, self, **kwargs))


class TreePopulation(ObjectIterableDataFrame):
    schema = DataFrameSchema(
        columns={
            **TREE_SCHEMA_COLS,
            **{
                "X": Column(
                    float,
                    nullable=True,
                    required=False,
                    title="X Coordinate (m)",
                    description="X coordinate of the tree in a projected coordinate system",
                ),
                "Y": Column(
                    float,
                    nullable=True,
                    required=False,
                    title="Y Coordinate (m)",
                    description="Y coordinate of the tree in a projected coordinate system",
                ),
            },
        },
        coerce=True,
        add_missing_columns=True,
        index=Index(int, unique=True),
    )

    def _row_to_object(self, row) -> Tree:
        """
        Convert a row of the dataframe to an object.
        """
        return Tree(
            species_code=row.SPCD,
            status_code=row.STATUSCD,
            diameter=row.DIA,
            height=row.HT,
            crown_ratio=row.CR,
            x=row.X,
            y=row.Y,
        )

    def apply_treatment(
        self, treatment: TreatmentProtocol | list[TreatmentProtocol]
    ) -> TreePopulation:
        """
        Applies one or more preconfigured silvicultural treatments to the tree population.

        Parameters
        ----------
        treatment : TreatmentProtocol | list[TreatmentProtocol]
            A single treatment or a list of treatments to apply.

        Returns
        -------
        TreePopulation
            A new TreePopulation object post treatment.
        """
        if isinstance(treatment, list):
            df = self.data.copy()
            for t in treatment:
                df = t.apply(df)
            return TreePopulation(df)
        else:
            return TreePopulation(treatment.apply(self.data))


class Tree:
    """
    An object representing an individual tree. The tree has required attributes
    that represent measurements of the tree, provide estimates of the tree's
    biomass and crown profile, and the tree's location in a projected
    coordinate system.

    Attributes
    ----------
    species_code : int
        An FIA integer species code.
    status_code : int
        1 = Live, 2 = Dead, 3 = Missing
    diameter : float
        Diameter (cm) of the tree measured at breast height (1.37 m). Units: cm.
    height : float
        Height (m) of the tree measured from the ground. Units: m.
    crown_ratio : float
        Ratio (-) of the crown length to the total tree height. Units: none.
    x : float
        X coordinate of the tree in a projected coordinate system. Units: m.
    y : float
        Y coordinate of the tree in a projected coordinate system. Units: m.
    """

    species_code: int
    status_code: int
    diameter: float
    height: float
    crown_ratio: float
    x: float
    y: float

    def __init__(
        self,
        species_code: int,
        status_code: int,
        diameter: float,
        height: float,
        crown_ratio: float,
        x=0,
        y=0,
        crown_profile_model_type="purves",
        biomass_allometry_model_type="NSVB",
    ):
        # TODO: Species code needs to be valid
        self.species_code = int(species_code)

        # TODO: Status code needs to be valid
        self.status_code = status_code

        # TODO: Diameter and height need to be greater than 0
        self.diameter = diameter
        self.height = height

        # TODO: Crown ratio needs to be between 0 and 1
        self.crown_ratio = crown_ratio

        self.x = x
        self.y = y

        if crown_profile_model_type not in ["beta", "purves"]:
            raise ValueError(
                "The crown profile model must be one of the following: 'beta' or 'purves'"
            )
        self._crown_profile_model_type = crown_profile_model_type

        available_biomass_allometry_models = ["NSVB", "jenkins"]
        if biomass_allometry_model_type == "NSVB" and self.species_group == 10:
            biomass_allometry_model_type = "jenkins"
        if biomass_allometry_model_type not in available_biomass_allometry_models:
            raise ValueError(
                f"Selected biomass allometry model: {biomass_allometry_model_type}. The biomass allometry model must be one of the following: {available_biomass_allometry_models}"
            )
        self._biomass_allometry_model_type = biomass_allometry_model_type

    @property
    def crown_length(self) -> float:
        """
        Returns the length of the tree's crown.
        """
        return self.height * self.crown_ratio

    @property
    def crown_base_height(self) -> float:
        """
        Returns the height at which the live crown starts.
        """
        return self.height - self.crown_length

    @property
    def species_group(self) -> int:
        """
        Returns the species group of the tree based on the species code.
        """
        return SPCD_PARAMS[str(self.species_code)]["SPGRP"]

    @property
    def is_live(self):
        """
        Returns True if the tree is alive, False otherwise. The Tree is alive if
        the status code is 1.
        """
        return self.status_code == 1

    @property
    def crown_profile_model(self) -> CrownProfileModel:
        if self._crown_profile_model_type == "beta":
            return BetaCrownProfile(
                self.species_code, self.crown_base_height, self.crown_length
            )
        elif self._crown_profile_model_type == "purves":
            return PurvesCrownProfile(
                self.species_code, self.diameter, self.height, self.crown_ratio
            )

    @property
    def max_crown_radius(self) -> float:
        """
        Returns the maximum radius of the tree's crown.
        """
        return self.crown_profile_model.get_max_radius()

    def get_crown_radius_at_height(self, height: float | ndarray) -> float | ndarray:
        """
        Uses the crown profile model to get the radius of the crown at a given
        height in meters.
        """
        return self.crown_profile_model.get_radius_at_height(height)

    @property
    def biomass_allometry_model(self) -> BiomassAllometryModel:
        if self._biomass_allometry_model_type == "jenkins":
            return JenkinsBiomassEquations(self.species_code, self.diameter)
        elif self._biomass_allometry_model_type == "NSVB":
            return NSVBEquations(
                self.species_code,
                self.diameter,
                self.height,
            )

    @property
    def foliage_biomass(self) -> float:
        """
        Returns the estimated foliage biomass of the tree
        """
        return self.biomass_allometry_model.estimate_foliage_biomass()

    def voxelize(
        self, horizontal_resolution: float, vertical_resolution: float, **kwargs
    ) -> VoxelizedTree:
        return voxelize_tree(self, horizontal_resolution, vertical_resolution, **kwargs)

    @classmethod
    def from_row(cls, row):
        return cls(
            species_code=int(row.SPCD),
            status_code=int(row.STATUSCD),
            diameter=row.DIA,
            height=row.HT,
            crown_ratio=row.CR,
            x=row.X,
            y=row.Y,
        )


class CrownProfileModel(ABC):
    """
    Abstract base class representing a tree crown profile model.
    The crown profile model is a rotational solid that can be queried at any
    height to get a radius, or crown width, at that height.

    This abstract class provides methods to get the radius of the crown at a
    given height and to get the maximum radius of the crown.
    """

    @abstractmethod
    def get_radius_at_height(self, height):
        """
        Abstract method to get the radius of the crown at a given height.
        """
        raise NotImplementedError(
            "The get_radius method must be implemented in a subclass."
        )

    @abstractmethod
    def get_max_radius(self) -> float:
        """
        Abstract method to get the maximum radius of the crown.
        """
        raise NotImplementedError(
            "The get_max_radius method must be implemented in a subclass."
        )


class BetaCrownProfile(CrownProfileModel):
    """
    A crown profile model based on a beta distribution.
    """

    a: float
    b: float
    c: float
    crown_length: float
    crown_base_height: float

    def __init__(
        self, species_code: int, crown_base_height: float, crown_length: float
    ):
        """
        Initializes a BetaCrownProfile instance.
        """
        self.species_code = int(species_code)
        self.crown_base_height = crown_base_height
        self.crown_length = crown_length

        species_group = SPCD_PARAMS[str(self.species_code)]["SPGRP"]
        self.a = SPGRP_PARAMS[str(species_group)]["BETA_CANOPY_a"]
        self.b = SPGRP_PARAMS[str(species_group)]["BETA_CANOPY_b"]
        self.c = SPGRP_PARAMS[str(species_group)]["BETA_CANOPY_c"]
        self.beta = beta(self.a, self.b)

    def get_max_radius(self) -> float:
        """
        Returns the maximum radius of the crown. This function finds the mode
        of the beta distribution, which is the value of z at which the beta
        distribution is at its max, and then calculates the radius at that
        height.
        """
        # Find the mode of the beta distribution. This is the value of z at
        # which the beta distribution is at its max.
        z_max = (self.a - 1) / (self.a + self.b - 2)
        normalized_max_radius = self._get_radius_at_normalized_height(z_max)
        return normalized_max_radius * self.crown_length

    def get_radius_at_height(self, height):
        """
        Returns the radius of the crown at a given height using the beta
        distribution crown model described in equation (3) in Ferrarese et
        al. (2015). Equation (3) gives the radius of the crown at a given
        height as a proportion of the crown length scaled between 0 and 1. To
        get a crown radius in meters, the result is multiplied by the crown
        length.
        """
        normalized_height = self._get_normalized_height(height)
        radius_at_normalized_height = self._get_radius_at_normalized_height(
            normalized_height
        )
        return radius_at_normalized_height * self.crown_length

    def _get_normalized_height(self, height):
        """
        Converts a height (in meters) of the tree crown to a unitless height
        between 0 and 1 representing the proportion of the crown length.
        """
        return (height - self.crown_base_height) / self.crown_length

    def _get_radius_at_normalized_height(self, z):
        """
        Returns the unitless radius of the crown at a given normalized height.
        The radius is scaled between 0 and 1 and represents a proportion of
        the crown length of the tree.

        The radius is calculated using the beta distribution probability
        density function (PDF) at the given normalized height. The PDF of the
        beta distribution uses an additional scaling factor, 'c', for the
        application to crown profiles and is described by equation (3) in
        Ferrarese et al. (2015).

        """
        z = np.asarray(z)
        mask = (z >= 0) & (z <= 1)

        result = np.zeros_like(z)
        result[mask] = (
            self.c * z[mask] ** (self.a - 1) * (1 - z[mask]) ** (self.b - 1) / self.beta
        )

        if result.size == 1:
            return result.item()  # Return as a scalar
        else:
            return result  # Return as an array


class PurvesCrownProfile(CrownProfileModel):
    """
    Purves Crown Profile Model.

    This class computes the crown profile for a given tree based on the Purves et al. (2007) model.
    It uses species-specific parameters to calculate the crown radius at different heights; parameters
    can be found in fastfuels_core/data/spcd_parameters.json.

    Parameters
    ----------
    species_code : int
        The species code of the tree.
    dbh : float
        Diameter at breast height (cm).
    height : float
        Total height of the tree (m).
    crown_ratio : float
        Ratio of the crown length to the total height of the tree.

    Attributes
    ----------
    height : float
        Total height of the tree (m).
    crown_ratio : float
        Ratio of the crown length to the total height of the tree.
    crown_base_height : float
        Height at which the crown starts (m).
    max_crown_radius : float
        Maximum radius of the crown (m).
    shape_parameter : float
        Shape parameter for the crown profile.

    Methods
    -------
    get_radius_at_height(height)
        Computes the crown radius at a given height.
    get_max_radius()
        Returns the maximum crown radius.
    """

    # See Purves et al. (2007) Table S2 in Supporting Information
    C0_R0 = 0.503
    C1_R0 = 3.126
    C0_R40 = 0.5
    C1_R40 = 10.0
    C0_B = 0.196
    C1_B = 0.511

    def __init__(
        self, species_code: int, dbh: float, height: float, crown_ratio: float
    ):
        """
        Initializes the PurvesCrownProfile model with the given parameters.

        Parameters
        ----------
        species_code : int
            The species code of the tree.
        dbh : float
            Diameter at breast height (cm).
        height : float
            Total height of the tree (m).
        crown_ratio : float
            Ratio of the crown length to the total height of the tree.
        """
        self.height = height
        self.crown_ratio = crown_ratio
        self.crown_base_height = height - height * crown_ratio

        trait_score = SPCD_PARAMS[str(species_code)]["PURVES_TRAIT_SCORE"]

        # Compute maximum crown radius
        r0j = (1 - trait_score) * self.C0_R0 + trait_score * self.C1_R0
        r40j = (1 - trait_score) * self.C0_R40 + trait_score * self.C1_R40
        self.max_crown_radius = r0j + (r40j - r0j) * (dbh / 40.0)

        # Compute crown shape parameter
        self.shape_parameter = (1 - trait_score) * self.C0_B + trait_score * self.C1_B

    def get_radius_at_height(self, height: float | np.ndarray) -> float | np.ndarray:
        """
        Computes the crown radius at a given height.

        Parameters
        ----------
        height : float or np.ndarray
            Height(s) at which to compute the crown radius (m).

        Returns
        -------
        float or np.ndarray
            Crown radius (m) at the given height(s). Returns a float if input height is a scalar,
            otherwise returns a numpy array.
        """
        height = np.asarray(height)
        radius = np.where(
            height < self.crown_base_height,
            0.0,
            self.max_crown_radius
            * (np.minimum((self.height - height) / self.height, 0.95) / 0.95)
            ** self.shape_parameter,
        )
        radius = np.nan_to_num(radius, nan=0.0)
        return radius if radius.size > 1 else radius.item()

    def get_max_radius(self) -> float | np.ndarray:
        """
        Returns the maximum crown radius.

        Returns
        -------
        float or np.ndarray
            Maximum crown radius (m).
        """
        return self.get_radius_at_height(self.crown_base_height)


class BiomassAllometryModel(ABC):
    """
    Abstract base class representing a tree biomass allometry model.
    The biomass allometry model is used to estimate the biomass of a tree
    based on one or more independent variables, such as diameter and height.
    """

    @abstractmethod
    def estimate_foliage_biomass(self) -> float:
        """
        Abstract method to estimate the biomass of a tree based on its diameter
        and height.
        """


class NSVBEquations(BiomassAllometryModel):
    """
    This class implements the National Scale Volume and Biomass Estimators
    modeling system.

    Information on the NSVB system can be found in the following publication:
    https://www.fs.usda.gov/research/treesearch/66998
    """

    def __init__(
        self,
        species_code: int,
        diameter: float,
        height: float,
        division: str = "",
        cull: int = 0,
        decay_code: int = 0,
        actual_height: float = None,
    ):
        if not _is_valid_spcd(species_code):
            raise ValueError(f"Species code {species_code} is not valid.")

        if not _is_valid_diameter(diameter):
            raise ValueError(f"Diameter {diameter} must be greater than 0.")

        self.species_code = species_code
        self.diameter = diameter
        self.height = height
        self.division = division
        self.cull = cull
        self.decay_code = decay_code
        self.actual_height = actual_height if actual_height else height

    def estimate_foliage_biomass(self):
        """
        Foliage weight is estimated by combining foliage weight coefficients
        and the appropriate model in Table S9a of the NSVB GTR.
        """
        dia_ft = self.diameter / 2.54
        height_ft = self.height * 3.28084
        dry_foliage_lb = total_foliage_dry_weight(
            spcd=self.species_code,
            dia=dia_ft,
            ht=height_ft,
            division=self.division,
        )
        return dry_foliage_lb * 0.453592  # Convert to kg


class JenkinsBiomassEquations(BiomassAllometryModel):
    """
    This class implements the National-Scale Biomass Estimators developed by
    Jenkins et al. (2003). These equations are used to estimate above ground and
    component biomass for 10 species groups in the United States.
    """

    def __init__(self, species_code, diameter):
        if not _is_valid_spcd(species_code):
            raise ValueError(f"Species code {species_code} is not valid.")

        if not _is_valid_diameter(diameter):
            raise ValueError(f"Diameter {diameter} must be greater than 0.")

        self.species_code = species_code
        self.diameter = diameter

        self._species_group = SPCD_PARAMS[str(species_code)]["SPGRP"]
        self._is_softwood = SPCD_PARAMS[str(species_code)]["CLASS"]
        self._sapling_adjustment = SPCD_PARAMS[str(species_code)]["SAPADJ"]

    def _estimate_above_ground_biomass(self):
        """
        Uses Equation 1 and parameters in Table 4 of Jenkins et al. 2003 to
        estimate the above ground biomass of a tree based on species group and
        diameter at breast height.

        Biomass equation: bm = Exp(β_0 + (β_1 * ln(dbh))) Where bm is above
        ground biomass (kg) for trees 2.5cm dbh and larger, dbh is the
        diameter at breast height (cm), and β_0 and β_1 are parameters
        estimated from the data.

        NOTE: This method also applies a sapling adjustment factor (sapadj) for
        trees with dbh <= 12.7cm (5 inches).
        """
        # Get the parameters for the species group
        beta_0 = SPGRP_PARAMS[str(self._species_group)]["JENKINS_AGB_b0"]
        beta_1 = SPGRP_PARAMS[str(self._species_group)]["JENKINS_AGB_b1"]

        # Estimate the above ground biomass
        biomass = np.exp(beta_0 + (beta_1 * np.log(self.diameter)))

        # Apply the sapling adjustment factor
        if self.diameter <= 12.7 and self._sapling_adjustment > 0:
            biomass *= self._sapling_adjustment

        return biomass

    def estimate_foliage_biomass(self):
        """
        Uses Equation 2 and parameters in Table 6 of Jenkins et al. 2003 to
        estimate component ratio and foliage biomass of a tree based on species
        group, hardwood classification, and diameter at breast height.

        Biomass ratio equation:
        r = Exp(β_0 + (β_1 / dbh))
        Where r is the ratio of component to total aboveground biomass for
        trees 2.5cm dbh and larger, dbh is the diameter at breast height in cm,
        and β_0 and β_1 are parameters estimated from the data.

        r is multiplied by the above ground biomass to estimate the
        foliage biomass in kg.
        """
        # Get the parameters for the hardwood class
        beta_0 = CLASS_PARAMS[str(self._is_softwood)]["foliage_b0"]
        beta_1 = CLASS_PARAMS[str(self._is_softwood)]["foliage_b1"]

        # Estimate the foliage component ratio
        ratio = np.exp(beta_0 + (beta_1 / self.diameter))

        # Estimate the foliage biomass
        foliage_biomass = self._estimate_above_ground_biomass() * ratio

        return foliage_biomass


def _is_valid_spcd(spcd: int) -> bool:
    return str(int(spcd)) in SPCD_PARAMS


def _is_valid_diameter(diameter: float) -> bool:
    return diameter > 0


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # Create subplots for different crown ratios
    crown_ratios = [0.2, 0.4, 0.6, 0.8]
    fig, axes = plt.subplots(1, len(crown_ratios), figsize=(20, 10), sharey=True)

    for ax, crown_ratio in zip(axes, crown_ratios):
        tree_beta = Tree(
            814, 1, 15.0, 5.0, crown_ratio, 0, 0, crown_profile_model_type="beta"
        )
        tree_purves = Tree(
            814, 1, 15.0, 5.0, crown_ratio, 0, 0, crown_profile_model_type="purves"
        )

        heights = np.linspace(0, tree_beta.height, 100)
        beta_p = tree_beta.get_crown_radius_at_height(heights)
        purves_p = tree_purves.get_crown_radius_at_height(heights)

        ax.plot(beta_p, heights, "-r", label="Beta Crown Profile")
        ax.plot(purves_p, heights, "-k", label="Purves Crown Profile")
        ax.set_xlabel("Crown Radius (m)")
        ax.set_title(f"Crown Ratio = {crown_ratio}")
        ax.set_aspect("equal")
        ax.legend()

    axes[0].set_ylabel("Height (m)")
    plt.gca().set_aspect("equal", adjustable="box")
    plt.show()



import json
import urllib.request
from setuptools import find_packages, setup


def read_file(fname):
    with open(fname, encoding="utf-8") as fd:
        return fd.read()


def get_requirements(fname):
    with open(fname, encoding="utf-8") as fd:
        reqs = [line.strip() for line in fd if line]
    return reqs


def get_version():
    """Get the version number."""
    with urllib.request.urlopen(
        "https://api.github.com/repos/silvxlabs/fastfuels-core/releases/latest"
    ) as response:
        data = json.loads(response.read().decode("utf-8"))
    version = data["tag_name"]
    return version[1:]  # Remove the leading "v" from the version number


NAME = "fastfuels-core"
DESCRIPTION = "Core algorithms for the FastFuels project"
LONG_DESCRIPTION = read_file("README.md")
VERSION = get_version()
LICENSE = "MIT"
URL = "https://github.com/silvxlabs/fastfuels-core"
PROJECT_URLS = {"Bug Tracker": f"{URL}/issues"}
INSTALL_REQUIRES = [
    "dask",
    "dask-expr",
    "fastparquet",
    "geopandas",
    "nsvb",
    "numpy>=2",
    "pandas",
    "pandera",
    "rasterio",
    "scipy",
    "shapely>=2",
]

EXTRAS_REQUIRE = {
    "plotting": [
        "matplotlib",
        "pyvista",
    ],
    "test": [
        "pytest",
        "matplotlib",
    ],
}
setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license=LICENSE,
    url=URL,
    project_urls=PROJECT_URLS,
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
    ],
    package_dir={"": "."},
    packages=find_packages(exclude=["docs", "tests"]),
    package_data={"fastfuels_core": ["data/*"]},
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    python_requires=">=3.9",
)
